"""Module for maintaining and defining user options"""
from .util import check_string


class ToggleOption(object):
    """An option that has no value, it is only either enabled or disabled. It
    can be specifically set to either enabled or disabled, or be toggled back
    and fourth.
    """

    def __init__(self, name, enabled_by_default):
        """A new ToggleOption that must have a name and a default setting"""
        self.name = name
        if enabled_by_default:
            self.__enabled = True
        else:
            self.__enabled = False

    def is_enabled(self):
        """Returns True if this option is enabled, else false"""
        return self.__enabled

    def enable_option(self):
        """Set option to be enabled"""
        self.__enabled = True

    def disable_option(self):
        """Set option to be disabled"""
        self.__enabled = False

    def toggle_option(self):
        """Toggle a "toggle" option between disabled and enabled"""
        if self.is_enabled():
            self.disable_option()
        else:
            self.enable_option()


class FreeFormTextOption(object):
    """An option that has a free form text value. These options will mostly be
    used for config file values such as format strings or complex user
    preferences.
    """

    def __init__(self, name, default_value):
        """A new FreeFormTextOption that must have a name and a default value
        """
        self.name = name
        self.__value = default_value

    def get_free_text_option(self):
        """Return the free form value"""
        return self.__value

    def set_free_text_option(self, value):
        """Add <value> to option manager. Option must be a string"""
        check_string(value, 'Value must be a string!')
        self.__value = value


SHORTEN_FILEPATHS_OPT = 'shorten_filepaths'
SYNTAX_HIGHLIGHT_OPT = 'syntax_highlighting'
DEBUG_MODE_OPT = 'debug_mode'

OPTION_DEFAULTS = {
    SHORTEN_FILEPATHS_OPT: ToggleOption(SHORTEN_FILEPATHS_OPT, False),
    SYNTAX_HIGHLIGHT_OPT: ToggleOption(SYNTAX_HIGHLIGHT_OPT, True),
    DEBUG_MODE_OPT: ToggleOption(DEBUG_MODE_OPT, False),
}


def singleton(cls):
    """Decorator to only return one instance of a class object"""
    instances = {}

    def getinstance():
        """Return a new or existing instance of this class"""
        if cls not in instances:
            instances[cls] = cls()
        return instances[cls]
    return getinstance


@singleton
class UserOptionManager(object):
    """An object to manage all the user related options.
    There should only ever be one instance of this object (see singleton
    above).
    """

    def __init__(self):
        """Contruct a new UOM"""
        # Just bootstrap options from the defaults. As of right now, I'm not
        # worried about modifying the defaults dict object. The only case where
        # this would cause issues is if you wanted to go back and check what
        # the default is during runtime, which I don't have a use case for yet.
        # This can be changed to a deepcopy if such a use case arises.
        self.options = OPTION_DEFAULTS

    def get_option(self, option):
        """Return the option object for the option name <option>"""
        # Option names are strings
        check_string(option, 'Option name must be a string!')
        # Just let python dict KeyError be raised if that option doesn't exist
        return self.options[option]



"""Screen object reacts to user input updating the curses screen"""
import curses
import logging

import texteditpad

from . import modes
from . import search
from . import git
from . import content as content_mod
from .options import UserOptionManager, SYNTAX_HIGHLIGHT_OPT

from .util import BidirectionalCycle


class BadModeException(Exception):
    """Simple exception to make exception handling easier up the stack"""
    pass


class UserError(Exception):
    """Simple exception to make exception handling easier up the stack"""
    pass


class CurrentPos(object):
    """A small data class to represent the positional data of a bblame screen.
    Attributes:
        cursor_line - The current line the cursor is on. Relative to the
                      visible screen, E.g.: cursor_line is always between
                      0 and screen height
        current_line - The current line of content, zero indexed from the top.
                       E.g. After freshly opening a blame, current_line is 0,
                            after scrolling down 5 lines, current_line is 4
        current_width - similar to current
    """
    def __init__(self, cursor_line, current_line, current_width):
        self.cursor_line = cursor_line
        self.current_line = current_line
        self.current_width = current_width

    def incr_cursor_line(self, num_lines=1):
        """Increment cursor line by <num_lines>"""
        self.cursor_line = self.cursor_line + num_lines

    def incr_current_line(self, num_lines=1):
        """Increment current line by <num_lines>"""
        self.current_line = self.current_line + num_lines

    def incr_current_width(self, num_cols=1):
        """Increment current width by <num_cols>"""
        self.current_width = self.current_width + num_cols

    def decr_cursor_line(self, num_lines=1):
        """Decrements cursor line by <num_lines>"""
        self.cursor_line = self.cursor_line - num_lines

    def decr_current_line(self, num_lines=1):
        """Decrements current line by <num_lines>"""
        self.current_line = self.current_line - num_lines

    def decr_current_width(self, num_cols=1):
        """Decrements current width by <num_cols>"""
        self.current_width = self.current_width - num_cols


class Screen(object):
    """The main class that bridges between the curses screen object and one
    or more blame or git show objects. Containing functions that are triggered
    by user interaction with the application which modify and/or read
    screen/git state.

    Attributes:
        stdscr - The curses screen object
        search_locs_iter - bidirectional iterator of the indexes to lines that
                           contain the current search pattern
        content_stack - A stack to keep track of git blame/show objects as
                        you browse file commit history and show commits.
        mode - The current mode bblame is in, see modes.py for more details
        search_str - current search string, if any.
    """

    def __init__(self, stdscr, filename, revision,
                 syntax_highlighting_enabled):
        self.current_pos = CurrentPos(0, 0, 0)
        self.search_locs_iter = None
        self.status_bar_next_msg = ''
        self.mode = modes.MODE_NORMAL
        self.content_stack = content_mod.ContentStack()
        self.gitlog = git.GitLog(filename)
        self.search_str = None
        self.syntax_highlighting_enabled = syntax_highlighting_enabled

        self.resize_windows(stdscr)  # prepares the subwindows
        self.set_status_bar_next_msg('')
        self._add_content_to_stack(git.Blame(filename,
                                             syntax_highlighting_enabled,
                                             revision[:git.ABBREV_LEN]))

    @staticmethod
    def setup_screen(scr):
        """setup screens
        set scrollingok and idlok for scrolling)
        set keypad for use of arrow keys
        """
        scr.scrollok(1)
        scr.idlok(1)
        scr.keypad(True)

    def resize_windows(self, app_scr):
        """Partition stdscr into properly sized subwindows"""
        height, width = app_scr.getmaxyx()
        self.stdscr = app_scr.derwin(height-1, width, 0, 0)
        self.status_bar = app_scr.derwin(1, width, height-1, 0)
        self.setup_screen(self.stdscr)
        self.setup_screen(self.status_bar)
        self.setup_screen(app_scr)

    def erase_status_bar(self):
        """Erase the screen"""
        self.status_bar.erase()

    def refresh_status_bar(self):
        """Refresh the statusbar"""
        self.status_bar.refresh()

    def getch(self, nodelay=False):
        """wait for char with delay or not"""
        self.stdscr.nodelay(nodelay)
        char = self.stdscr.getch()
        self.stdscr.nodelay(False)
        return char

    def check_for_escape(self, char):
        """Check if the escape key was pressed, which is tricky as the Alt
        key combos also use escape"""
        if char == 27:
            # At this point we don't know if it was ESC or an ALT key combo
            # Grab another key to see if it was an ALT combo
            next_char = self.getch(nodelay=True)
            esc = False
            if next_char == -1:
                # Escape was pressed
                esc = True
            # NOTE: Don't care about ALT right now so just drop it on the floor
            #       if it was.
            return esc

    def _get_max_yx(self):
        """return the max width and height of the screen"""
        return self.stdscr.getmaxyx()

    def _get_visible_content_len(self):
        """Get the length of visible content, I.e. the number of lines of
        blame or git show that are visible to the user"""
        screen_h, screen_w = self._get_max_yx()
        scr_content = self.get_current_scr_content()
        return min(len(scr_content) - self.current_pos.current_line, screen_h)

    def add_head_blame(self):
        """Add blame for the most recent commit of this file"""
        self._add_blame(self.gitlog.head)

    def add_tail_blame(self):
        """Add blame for the first commit of this file"""
        self._add_blame(self.gitlog.tail)

    def add_blame_parent(self):
        """Add blame of the parent of the current rev to the content stack"""
        curr_rev = self.get_current_scr_content().git_sha
        git_log_entry = self.gitlog[curr_rev]
        self._add_blame(git_log_entry.parent)

    def add_blame_ancestor(self):
        """Add blame of the ancestor of the current rev to the content stack"""
        curr_rev = self.get_current_scr_content().git_sha
        git_log_entry = self.gitlog[curr_rev]
        self._add_blame(git_log_entry.ancestor)

    def add_blame_drill(self):
        """Add a blame to the content stack for the revision under
        the cursor"""
        _, line_meta = self._get_cursor_line(include_meta=True)
        rev_to_drill = line_meta.sha
        if rev_to_drill == '0'*len(rev_to_drill):
            # This is likely a non-committed change. The parent of a
            # non-committed change is HEAD of the file
            head_sha = self.get_current_scr_content().git_sha
            git_log_entry = self.gitlog[head_sha]
            self._add_blame(git_log_entry.parent)
        elif rev_to_drill.startswith('^'):
            # This is a boundary commit, can't drill past this
            raise UserError('Can\'t drill past boundary commit, no parent')
        else:
            git_log_entry = self.gitlog[rev_to_drill]
            self._add_blame(git_log_entry.parent)

    def _add_blame(self, git_log_entry):
        """Add a blame to the content stack. Takes the file we're blaming and
        a function that when called returns the revision of the file we're
        blaming"""
        try:
            if git_log_entry is None:
                logging.warning('YOU HAVE REACHED THE BOTTOM OF THE WELL!')
                current_rev = self.get_current_scr_content().git_sha
                raise UserError('Error: No commit before %s' % current_rev)
            filename = git_log_entry.filename
            revision = git_log_entry.sha
            newblame = git.Blame(filename, self.syntax_highlighting_enabled,
                                 '%s' % revision)
        except git.BadRevException:
            logging.warning('BAD REV SELECTED! %s', revision)
            raise UserError('Error: bad revision (%s) selected', revision)
        self._add_content_to_stack(newblame)
        self.update_search_locs()
        self.mode = modes.MODE_NORMAL

    def get_current_scr_content(self):
        """return the current screen content"""
        stackframe = self.content_stack.peek()
        return stackframe.content

    def _add_content_to_stack(self, content):
        """Add content to the content_stack, along with the current state
        needed to restore the screen after we remove it"""
        self.content_stack.add(content, self.current_pos.current_line,
                               self.current_pos.current_width,
                               self.current_pos.cursor_line, self.mode)

    def restore_prev_content_and_state(self):
        """Delete the current frame of the content stack, restore the
        snapshot of the previous screen state it stored"""
        try:
            stackframe = self.content_stack.pop()
        except IndexError:
            raise UserError('No content to pop')

        self.current_pos.current_line = stackframe.last_current_line
        self.current_pos.current_width = stackframe.last_current_width
        self.current_pos.cursor_line = stackframe.last_cursor_line
        self.mode = stackframe.last_mode

        self.update_search_locs()

    def display_help(self, help_obj):
        """Display help"""
        self._add_content_to_stack(help_obj)
        self.mode = modes.MODE_HELP

    def _get_line(self, line, include_meta=False):
        """Return the content at index <line> from the current content being
        displayed and the attributes of that line"""
        scr_content = self.get_current_scr_content()
        if include_meta:
            meta_line = scr_content.lines_metadata[line]
            return (scr_content[line], meta_line)
        return scr_content[line]

    def _get_cursor_line(self, include_meta=False):
        """Return the line that the cursory is on. If <include_meta>
        then also return the meta_data for the line, most commonly this is used
        to get the git sha for the line"""
        cursline_2_bline = (self.current_pos.cursor_line +
                            self.current_pos.current_line - 1)
        return self._get_line(cursline_2_bline, include_meta=include_meta)

    def set_status_bar_next_msg(self, content,
                                attributes=curses.A_DIM | curses.A_STANDOUT):
        """Set a message to be displayed in the status bar on the next screen
        refresh. Returns a tuple with the message content and any
        attributes"""
        self.status_bar_next_msg = (content, attributes)

    def status_bar_toast(self, content):
        """Set a message to be displayed in the status bar on the next screen
        refresh with bold attribute"""
        self.status_bar_next_msg = (content, curses.A_BOLD | curses.A_STANDOUT)

    def get_status_bar_content(self):
        """Return content to be displayed in the status bar. A standard
        template for each mode will be displayed unless an explicit status
        bar messages was set to be displayed"""
        scr_content = self.get_current_scr_content()
        height, width = self._get_max_yx()
        ret_msg = ''
        ret_attrs = curses.A_DIM | curses.A_STANDOUT

        if self.status_bar_next_msg[0]:
            ret_msg, ret_attrs = self.status_bar_next_msg
            self.set_status_bar_next_msg("")
        elif self.mode == modes.MODE_HELP:
            ret_msg = 'Help'
        else:
            git_sha = scr_content.git_sha
            if self.mode == modes.MODE_SHOW:
                ret_msg = '%s: %s' % ('show', git_sha)
            else:
                filename = scr_content.filename
                ret_msg = '%s: %s (%s)' % ('blame', filename, git_sha)
        return ret_msg, ret_attrs

    def _draw_status_bar(self):
        """Fetch the correct content to be displayed in the status bar and
        draw it in the status bar window."""
        content, attributes = self.get_status_bar_content()
        _, width = self._get_max_yx()
        content.ljust(width-1)
        # If the message is too long for the width of the screen, truncate it
        # and add an ellipses to indicate we've done so.
        if len(content) > width - 1:
            content = content[:width-4] + '...'

        self.erase_status_bar()
        self.status_bar.addstr(content[:width-1], attributes)
        self.refresh_status_bar()

    def init_git_show_file(self):
        """Setup for a git show for the file commit"""
        git_sha = self.get_current_scr_content().git_sha
        self.init_git_show(git_sha)

    def init_git_show(self, rev_to_show=None):
        """Setup for a git show"""
        rev = rev_to_show
        if not rev:
            _, line_meta = self._get_cursor_line(include_meta=True)
            rev = line_meta.sha

        if rev == '0'*len(rev):
            uncommitted_show_msg = 'Cannot git show an uncommitted change!'
            logging.debug(uncommitted_show_msg)
            self.status_bar_toast(uncommitted_show_msg)
        else:
            show = git.Show(rev)
            self._add_content_to_stack(show)
            self.mode = modes.MODE_SHOW
            # Snap to top of git show
            self.current_pos.current_line = 0
            self.current_pos.current_width = 0
            self.update_search_locs()

    def get_search_string(self):
        """Get search string using texteditpad and some trickery"""
        # Create a subwindow in the statusbar to contain the single line
        # Textbox editor
        screen_h, screen_w = self._get_max_yx()
        # Create massive window to avoid hitting max width
        temp_search_win = curses.newwin(1, 900, screen_h, 1)
        txtbox = texteditpad.Textbox(temp_search_win, insert_mode=True)
        # Clear the status bar of status indicators, and print the search
        # string entry '/'
        self.erase_status_bar()
        self.status_bar.addstr(0, 0, '/')
        self.refresh_status_bar()
        # Enable the cursor and begin excepting text until the user hits return
        curses.curs_set(1)
        self.search_str = search.collect_search_string(txtbox)
        curses.curs_set(0)
        logging.info('SEARCH: search string: %s~', self.search_str)
        if not self.search_str:
            # Don't update anything since it will raise a user toast since no
            # matches will be found for the search string
            logging.info('SEARCH: no search string')
            return
        self.update_search_locs()
        self.jump_to_next_search_match()

    def jump_to_next_search_match(self):
        """Jump to the next search match"""
        self._jump_to_search_match_helper(search.SEARCH_STRING_JUMP_FORWARD)

    def jump_to_prev_search_match(self):
        """Jump to the previous search match"""
        self._jump_to_search_match_helper(search.SEARCH_STRING_JUMP_BACKWARD)

    def _jump_to_search_match_helper(self, direction):
        """Jump to next search match in <direction>"""
        if self.search_locs_iter is not None:
            try:
                if direction == search.SEARCH_STRING_JUMP_FORWARD:
                    self.current_pos.current_line = next(self.search_locs_iter)
                elif direction == search.SEARCH_STRING_JUMP_BACKWARD:
                    self.current_pos.current_line = \
                            self.search_locs_iter.prev()
                else:
                    raise Exception('Unknown search direction (%s) passed',
                                    direction)
                # Kick the user back to Normal/show mode when searching
                if self.mode == modes.MODE_VISUAL:
                    self.mode = modes.MODE_NORMAL
            except StopIteration:
                self.status_bar_toast('Error: Pattern not found')
        else:
            logging.info('No search pattern entered')

    def init_line_arg(self, content_line):
        """Setup the screen given the provided <content_line> from the user,
        passed from the command line.
        This is roughly:
            - centre the screen on that line if possible (stopping at top or
              bottom of content).
            - enable visual mode (placing the cursor on the location of the
              screen that is over overtop of <content_line>)."""
        screen_h, screen_w = self._get_max_yx()
        screen_loc_of_line = self._center_screen_on_line(content_line)
        self.init_vis_cursor(screen_loc_of_line)

    def init_vis_cursor(self, cursor_line=None):
        """init visual select cursor mode.
        Move to the correct mode and initialize the cursor to a either
        the provided line <cursor_line>, the position of the current search
        (if one exists and it's located in the current screen view) or the
        default (in the middle of the visible content)"""
        self.mode = modes.MODE_VISUAL
        screen_h, screen_w = self._get_max_yx()

        def current_search_line():
            """Turn search location into a one-based index content line"""
            return self.search_locs_iter.curr() + 1

        if cursor_line:
            self.current_pos.cursor_line = cursor_line
        # If we have a search and the current search hit is anywhere in the
        # viewable area
        elif (self.search_locs_iter and
              self.content_line_to_screen_line(current_search_line())):
            self.current_pos.cursor_line = self.content_line_to_screen_line(
                current_search_line())
        else:
            # If our blame has less lines than height then we want to init
            # the cursor to the middle of the blame content not into the middle
            # of the screen which could be off of the edge of file content.
            self.current_pos.cursor_line = self._get_visible_content_len()//2

    def content_line_to_screen_line(self, content_line):
        """Convert a content line (e.g. Line 4356 of a file/blame) and return
        the line on the viewable screen that <content_line> is on (e.g. screen
        line 6, i.e. six lines from the top of the screen the user sees).
        Returns None if <content_line> is not in the viewable area"""
        scrn_content_len = self._get_visible_content_len()
        current_line = self.current_pos.current_line
        if current_line < content_line <= current_line + scrn_content_len:
            return content_line - current_line

    def _center_screen_on_line(self, line):
        """Centers the screen around <line> of the blame as best as possible.
        Returns the location on the viewable screen space, where the line we
        wanted to center on ended up. In most cases this will indeed be the
        center (height/2) but may be elsewhere if we've reached the end of the
        git blame"""
        screen_h, screen_w = self._get_max_yx()
        scrn_content_len = len(self.get_current_scr_content())
        if line < screen_h//2:
            # Centering the screen at a line which would leave less than half
            # a screen height worth of content above wouldn't leave us with
            # enough content to fill the screen, so cap it at 0
            self.current_pos.current_line = 0
        elif line > (scrn_content_len - screen_h//2 - 1):
            # Same as above but we'd run off the bottom of the content, cap
            # things such that we have at least one screens worth of content
            # to display.
            self.current_pos.current_line = scrn_content_len - screen_h
        else:
            self.current_pos.current_line = line - screen_h//2
        return max(1, min(line - self.current_pos.current_line, screen_h))

    def _re_center_screen(self):
        """Called to recenter the screen.
        Redrawing the screen after drilling into a blame or popping back may
        leave all or part of the screen outside of the content, so recenter
        the screen on the new/changed content"""
        scrn_content = self.get_current_scr_content()
        screen_h, screen_w = self._get_max_yx()
        # recenter the screen if we're out of the content
        if self.current_pos.current_line > len(scrn_content):
            if len(scrn_content) < screen_h//2:
                self.current_pos.current_line = 0
            else:
                self.current_pos.current_line = len(scrn_content) - screen_h//2
        self._re_center_cursor()

    def _re_center_cursor(self):
        """Recenter the visual cursor"""
        screen_h, screen_w = self._get_max_yx()
        # recenter the visual mode cursor if it's outside the content
        if self.current_pos.cursor_line > self._get_visible_content_len():
            self.current_pos.cursor_line = self._get_visible_content_len()//2

    def _highlight_for_vis_cursor(self, content_index):
        """determines whether the current line should be highlighted for
        visual mode cursor"""
        if (self.mode == modes.MODE_VISUAL and
                content_index == (self.current_pos.cursor_line - 1)):
            return True
        else:
            return False

    def update_search_locs(self):
        """Creates a bi-directional iterator of all lines that contain the
        search pattern."""
        scr_content = self.get_current_scr_content()
        _search_locs = search.find_search_locs(scr_content, self.search_str)
        self.search_locs_iter = BidirectionalCycle(_search_locs)
        logging.info('SEARCH: Done updating search locs')

    def _draw_mapped_tokens(self, mapped_tokens, current_index):
        # TODO: possibly remove this
        """Draw line that is a list of tokens for search.
        This process is novel because the tokens have different curses attrs,
        so they must be added to the scr in pieces, and we also must be
        careful to stay within the current screen y position and width"""
        screen_h, screen_w = self._get_max_yx()
        curr_x_pos = start_pos = 0
        for token_str, highlight_for_search, token_attrs in mapped_tokens:
            token_len = len(token_str)
            if curr_x_pos == 0:
                # Until we've started printing the line check how much, if
                # any, of the current token fits in the screen width, it may
                # need to be clipped
                len_before_begin = ((start_pos + token_len) -
                                    self.current_pos.current_width)
                if len_before_begin < 0:
                    # Line not yet in the current width, (I.e. The user has
                    # shifted the screen over to the right, and the token we're
                    # currently processing doesn't fall in the viewable area
                    start_pos += token_len
                    continue
                # clip the token, as only part of it is in the viewable area
                token_str = token_str[token_len-len_before_begin:]
                token_len = len(token_str)
            # Check if the token will go past the width of the screen
            len_past_end = (curr_x_pos + token_len) - screen_w + 1
            if len_past_end > 0:
                # line would be longer than screen width, so trim
                token_str = token_str[0:-len_past_end]
            # If the current visual cursor line contains a search match, then
            # invert the highlight of the match so it still stands out
            if self._highlight_for_vis_cursor(current_index):
                highlight_for_search = not highlight_for_search
            if highlight_for_search:
                self.stdscr.addstr(current_index, curr_x_pos,
                                   token_str.encode('utf-8'),
                                   curses.A_STANDOUT | token_attrs)
            else:
                self.stdscr.addstr(current_index, curr_x_pos,
                                   token_str.encode('utf-8'), token_attrs)
            # Increment the x position by the length of the token we've added
            curr_x_pos += token_len
            if len_past_end > 0:
                break

    def redraw_screen(self):
        """Clear and redraw the screen, print each line and be sure to
        highlight the cursor line if we are in visual mode"""
        opt_mngr = UserOptionManager()
        syntax_toggle_opt = opt_mngr.get_option(SYNTAX_HIGHLIGHT_OPT)
        self._re_center_screen()
        screen_h, screen_w = self._get_max_yx()
        # <line_idx> is the index to the actual line of text from the file
        # that's being printed to the screen at line <scr_idx>. <scr_idx> is
        # always between 0 and height of the screen
        curr_line = self.current_pos.current_line
        curr_width = self.current_pos.current_width
        for scr_idx, line_idx in enumerate(range(curr_line,
                                                 curr_line + screen_h)):
            line = self._get_line(line_idx)
            logging.debug('drawing line: %s', line.line_segments)
            logging.debug('line: %s', str(line.full_text().encode('utf-8')))

            # Only print as much of the line that will fit in the screen width
            line = line.prepare_line_for_printing(curr_width,
                                                  curr_width+screen_w-1)
            logging.debug('len of new sub line: %d, width: %d',
                          len(line), curr_width)

            if self.search_locs_iter and line_idx in self.search_locs_iter:
                # This line contains at least one instance of the search string
                if self._highlight_for_vis_cursor(scr_idx):
                    line = line.highlight_str(self.search_str, None)
                else:
                    line = line.highlight_str(self.search_str,
                                              curses.A_STANDOUT)
                logging.info('highlighted line: %s', line.line_segments)

            num_chars_so_far = 0
            logging.debug('Printing full line: %s', line.full_text())
            for segment in line:
                logging.debug('printing segment: %s, with attr %s and '
                              'syntax %s, on screen line: %s',
                              segment.text, segment.curses_attrs,
                              segment.syntax_attrs, scr_idx)
                high_for_vis = self._highlight_for_vis_cursor(scr_idx)
                if high_for_vis and segment.text != self.search_str:
                    self.stdscr.addstr(scr_idx, num_chars_so_far,
                                       segment.text.encode('utf-8'),
                                       curses.A_STANDOUT | segment.curses_attrs
                                       )
                else:
                    if syntax_toggle_opt.is_enabled() and segment.syntax_attrs:
                        curses_attrs = segment.syntax_attrs
                        logging.debug('syntax highlighting!')
                    else:
                        # If not using syntax highlighting just use the default
                        # curses attribute for the segment
                        curses_attrs = segment.curses_attrs
                        logging.debug('NO syntax highlighting!')

                    logging.info('Segment attrs: %s', curses_attrs)
                    self.stdscr.addstr(scr_idx, num_chars_so_far,
                                       segment.text.encode('utf-8'),
                                       curses_attrs)

                num_chars_so_far += len(segment.text)

        self._draw_status_bar()

    def move_scr_down_page(self):
        """Move down by a page or less if there isn't a page full of content
        left"""
        scr_content = self.get_current_scr_content()
        screen_h, screen_w = self._get_max_yx()
        len_left = len(scr_content) - self.current_pos.current_line - screen_h
        if (len_left) > 0:
            page_size = min(len_left, screen_h)
            self.current_pos.incr_current_line(page_size)

    def move_scr_up_page(self):
        """move up by a page or less if there isn't a page full of content
        left"""
        screen_h, screen_w = self._get_max_yx()
        if self.current_pos.current_line > 0:
            page_size = min(screen_h, self.current_pos.current_line)
            self.current_pos.decr_current_line(page_size)

    def move_scr_right_nchars(self, nchars=8):
        """Move screen to the right by nchars"""
        screen_h, screen_w = self._get_max_yx()
        self.current_pos.incr_current_width(nchars)

    def move_scr_left_nchars(self, nchars=8):
        """Move screen to the left by nchars"""
        if self.current_pos.current_width > 0:
            self.current_pos.decr_current_width(nchars)

    def move_scr_down(self):
        """Move the screen down one line if we haven't reached the bottom yet
        """
        logging.info('MOVE_SCR_DOWN')
        scr_content = self.get_current_scr_content()
        screen_h, screen_w = self._get_max_yx()
        if (len(scr_content) - self.current_pos.current_line - screen_h) > 0:
            self.current_pos.incr_current_line()

    def move_scr_to_top(self):
        """Snap screen to top of content"""
        self.current_pos.current_line = 0

    def move_scr_to_bottom(self):
        """Snap screen to bottom of content"""
        scr_content = self.get_current_scr_content()
        screen_h, screen_w = self._get_max_yx()
        self.current_pos.current_line = len(scr_content) - screen_h

    def move_scr_up(self):
        """Move the screen up one lines if we haven't reached the top yet"""
        if self.current_pos.current_line > 0:
            self.current_pos.decr_current_line()

    def move_cursor_down(self):
        """Move the highlighted cursor line down by one line"""
        visable_content_len = self._get_visible_content_len()
        if self.current_pos.cursor_line <= visable_content_len:
            if self.current_pos.cursor_line == visable_content_len:
                # highlight is at the bottom of the screen
                self.move_scr_down()
            else:
                self.current_pos.incr_cursor_line()

    def move_cursor_up(self):
        """Move the highlighted cursor line up by one line"""
        if self.current_pos.cursor_line > 0:
            if self.current_pos.cursor_line == 1:
                # highlight is at the top of the screen
                self.move_scr_up()
            else:
                self.current_pos.decr_cursor_line()

    def log_state(self):
        """print some state to a log file.
        This is only printed to the logfile if the logging level is set
        appropriately. There is a user option that toggles the level of logging
        bblame starts with"""
        content = self.get_current_scr_content()
        screen_h, screen_w = self._get_max_yx()
        logging.info('Content stack len: %s', len(self.content_stack))
        logging.info('Mode: %s', self.mode)
        logging.info('Screen height: %s', screen_h)
        logging.info('Screen width: %s', screen_w)
        logging.info('Current line: %s', self.current_pos.current_line)
        logging.info('Current width: %s', self.current_pos.current_width)
        logging.info('Cursor line: %s', self.current_pos.cursor_line)
        if self.search_locs_iter:
            logging.info('Current search line: %s',
                         self.search_locs_iter.curr())
        logging.info('Length of file: %s', len(content))
        logging.info('Length of file exposed: %d',
                     self._get_visible_content_len())
        logging.info('length of file - currline %s',
                     (len(content) - self.current_pos.current_line))
        logging.info('range: %s\n',
                     str(range(self.current_pos.current_line,
                               self.current_pos.current_line + screen_h)))



"""Module for syntax highlighting file content using pygments"""
import logging
import curses
import re

# pylint: disable=no-name-in-module
from pygments.formatters import TerminalFormatter
from pygments.lexers import get_lexer_for_filename, guess_lexer
from pygments import highlight
from pygments.util import ClassNotFound


FORMATTER = TerminalFormatter()


# ANSI style codes
# Not all codes are used currently, but keep them all here for completeness
ANSI_RESET = 0
ANSI_BOLD_ON = 1
ANSI_BOLD_OFF = 22
ANSI_ITALICS_ON = 3
ANSI_ITALICS_OFF = 23
ANSI_UNDERLINE_ON = 4
ANSI_UNDERLINE_OFF = 24
ANSI_INVERSE_ON = 7
ANSI_INVERSE_OFF = 27
ANSI_STRIKETHROUGH_ON = 9
ANSI_STRIKETHROUGH_OFF = 29

# ANSI colour codes
ANSI_COLOR_BLACK_FG = 30
ANSI_COLOR_BLACK_BG = 40
ANSI_COLOR_RED_FG = 31
ANSI_COLOR_RED_BG = 41
ANSI_COLOR_GREEN_FG = 32
ANSI_COLOR_GREEN_BG = 42
ANSI_COLOR_YELLOW_FG = 33
ANSI_COLOR_YELLOW_BG = 43
ANSI_COLOR_BLUE_FG = 34
ANSI_COLOR_BLUE_BG = 44
ANSI_COLOR_MAGENTA_FG = 35
ANSI_COLOR_MAGENTA_BG = 45
ANSI_COLOR_CYAN_FG = 36
ANSI_COLOR_CYAN_BG = 46
ANSI_COLOR_WHITE_FG = 37
ANSI_COLOR_WHITE_BG = 47
ANSI_COLOR_DEFAULT_FG = 39
ANSI_COLOR_DEFAULT_BG = 49

# New "Bright" colours, no need for bold
ANSI_COLOR_BRIGHT_BLACK_FG = 90
ANSI_COLOR_BRIGHT_BLACK_BG = 100
ANSI_COLOR_BRIGHT_RED_FG = 91
ANSI_COLOR_BRIGHT_RED_BG = 101
ANSI_COLOR_BRIGHT_GREEN_FG = 92
ANSI_COLOR_BRIGHT_GREEN_BG = 102
ANSI_COLOR_BRIGHT_YELLOW_FG = 93
ANSI_COLOR_BRIGHT_YELLOW_BG = 103
ANSI_COLOR_BRIGHT_BLUE_FG = 94
ANSI_COLOR_BRIGHT_BLUE_BG = 104
ANSI_COLOR_BRIGHT_MAGENTA_FG = 95
ANSI_COLOR_BRIGHT_MAGENTA_BG = 105
ANSI_COLOR_BRIGHT_CYAN_FG = 96
ANSI_COLOR_BRIGHT_CYAN_BG = 106
ANSI_COLOR_BRIGHT_WHITE_FG = 97
ANSI_COLOR_BRIGHT_WHITE_BG = 107

ANSI_FOREGROUND = frozenset({
    ANSI_COLOR_BLACK_FG,
    ANSI_COLOR_RED_FG,
    ANSI_COLOR_GREEN_FG,
    ANSI_COLOR_YELLOW_FG,
    ANSI_COLOR_BLUE_FG,
    ANSI_COLOR_MAGENTA_FG,
    ANSI_COLOR_CYAN_FG,
    ANSI_COLOR_WHITE_FG,
    ANSI_COLOR_DEFAULT_FG,
    ANSI_COLOR_BRIGHT_BLACK_FG,
    ANSI_COLOR_BRIGHT_RED_FG,
    ANSI_COLOR_BRIGHT_GREEN_FG,
    ANSI_COLOR_BRIGHT_YELLOW_FG,
    ANSI_COLOR_BRIGHT_BLUE_FG,
    ANSI_COLOR_BRIGHT_MAGENTA_FG,
    ANSI_COLOR_BRIGHT_CYAN_FG,
    ANSI_COLOR_BRIGHT_WHITE_FG,
})

ANSI_BACKGROUND = frozenset({
    ANSI_COLOR_BLACK_BG,
    ANSI_COLOR_RED_BG,
    ANSI_COLOR_GREEN_BG,
    ANSI_COLOR_YELLOW_BG,
    ANSI_COLOR_BLUE_BG,
    ANSI_COLOR_MAGENTA_BG,
    ANSI_COLOR_CYAN_BG,
    ANSI_COLOR_WHITE_BG,
    ANSI_COLOR_DEFAULT_BG,
    ANSI_COLOR_BRIGHT_BLACK_BG,
    ANSI_COLOR_BRIGHT_RED_BG,
    ANSI_COLOR_BRIGHT_GREEN_BG,
    ANSI_COLOR_BRIGHT_YELLOW_BG,
    ANSI_COLOR_BRIGHT_BLUE_BG,
    ANSI_COLOR_BRIGHT_MAGENTA_BG,
    ANSI_COLOR_BRIGHT_CYAN_BG,
    ANSI_COLOR_BRIGHT_WHITE_BG,
})

ANSI_BRIGHT = frozenset({
    ANSI_COLOR_BRIGHT_BLACK_FG,
    ANSI_COLOR_BRIGHT_RED_FG,
    ANSI_COLOR_BRIGHT_GREEN_FG,
    ANSI_COLOR_BRIGHT_YELLOW_FG,
    ANSI_COLOR_BRIGHT_BLUE_FG,
    ANSI_COLOR_BRIGHT_MAGENTA_FG,
    ANSI_COLOR_BRIGHT_CYAN_FG,
    ANSI_COLOR_BRIGHT_WHITE_FG,
    ANSI_COLOR_BRIGHT_BLACK_BG,
    ANSI_COLOR_BRIGHT_RED_BG,
    ANSI_COLOR_BRIGHT_GREEN_BG,
    ANSI_COLOR_BRIGHT_YELLOW_BG,
    ANSI_COLOR_BRIGHT_BLUE_BG,
    ANSI_COLOR_BRIGHT_MAGENTA_BG,
    ANSI_COLOR_BRIGHT_CYAN_BG,
    ANSI_COLOR_BRIGHT_WHITE_BG,
})

ANSI_MODIFIERS = frozenset({
    ANSI_BOLD_ON,
    ANSI_ITALICS_ON,
    ANSI_UNDERLINE_ON,
    ANSI_INVERSE_ON,
})

# ANSI --> curses attributes map
ANSI_TO_CURSES = {
    ANSI_RESET: curses.A_NORMAL,
    ANSI_BOLD_ON: curses.A_BOLD,
    ANSI_UNDERLINE_ON: curses.A_UNDERLINE,
    ANSI_INVERSE_ON: curses.A_REVERSE,
    ANSI_COLOR_DEFAULT_FG: curses.A_NORMAL,
    ANSI_COLOR_DEFAULT_BG: curses.A_NORMAL,
    ANSI_COLOR_BLACK_FG: curses.COLOR_BLACK,
    ANSI_COLOR_RED_FG: curses.COLOR_RED,
    ANSI_COLOR_GREEN_FG: curses.COLOR_GREEN,
    ANSI_COLOR_YELLOW_FG: curses.COLOR_YELLOW,
    ANSI_COLOR_BLUE_FG: curses.COLOR_BLUE,
    ANSI_COLOR_MAGENTA_FG: curses.COLOR_MAGENTA,
    ANSI_COLOR_CYAN_FG: curses.COLOR_CYAN,
    ANSI_COLOR_WHITE_FG: curses.COLOR_WHITE,
    ANSI_COLOR_BRIGHT_BLACK_FG: curses.COLOR_BLACK,
    ANSI_COLOR_BRIGHT_RED_FG: curses.COLOR_RED,
    ANSI_COLOR_BRIGHT_GREEN_FG: curses.COLOR_GREEN,
    ANSI_COLOR_BRIGHT_YELLOW_FG: curses.COLOR_YELLOW,
    ANSI_COLOR_BRIGHT_BLUE_FG: curses.COLOR_BLUE,
    ANSI_COLOR_BRIGHT_MAGENTA_FG: curses.COLOR_MAGENTA,
    ANSI_COLOR_BRIGHT_CYAN_FG: curses.COLOR_CYAN,
    ANSI_COLOR_BRIGHT_WHITE_FG: curses.COLOR_WHITE,
}

# A_ITALIC was added in python3.7
try:
    curses_italics = curses.A_ITALIC
    ANSI_TO_CURSES[ANSI_ITALICS_ON] = curses.A_ITALIC
except AttributeError:
    pass


def get_lexer(lines, filename):
    """Given the lines of text for the file and filename, try find a lexer
    that matches, returning the lexer object or None"""
    text = '\n'.join(lines)
    lexer = None
    try:
        lexer = get_lexer_for_filename(filename)
    except ClassNotFound:
        try:
            # Guess the lexer by the content of the file
            lexer = guess_lexer(text)
        except ClassNotFound:
            logging.info('pygments couldn\'t determine lexer')

    return lexer


def highlight_lines(lines, lexer):
    """Take the lines of text to be highlighted and return a list of
    highlighted lines using pygments"""
    # Join into one string since pygments needs one string to properly
    # highlight multi line statements.
    text = '\n'.join(lines)
    highlighted_lines = highlight(text, lexer, FORMATTER).splitlines()
    # NOTE: pygments.highlight clobbers any blank lines at the beginning of the
    # file, count how many we should have and re add them.
    # Assemble a list of empty lists. Since parse_highlighed_line returns just
    # an empty list for an empty line.
    # Get empty lines at the beginning of the file
    empty_lines_start = []
    empty_lines_end = []
    curr_line_start = lines[0]
    curr_line_end = lines[-1]
    while '' in [curr_line_start, curr_line_end]:
        if curr_line_start == '':
            empty_lines_start.append([])
            curr_line_start = lines[len(empty_lines_start)]
        if curr_line_end == '':
            empty_lines_end.append([])
            curr_line_end = lines[-len(empty_lines_end)]

    return (empty_lines_start +
            # Parse each line into curses attributes
            [parse_highlighed_line(highlighted_line)
             for highlighted_line in highlighted_lines] +
            empty_lines_end)


def parse_highlighed_line(highlighted_line):
    """Takes an ANSI highlighted line of text and returns a list of tuples,
    each of which describing a section of the line and its associated curses
    attribute to colour it."""
    logging.debug('Working on line: %s:', highlighted_line)
    split_tokens = highlighted_line.split('\x1b[')
    saved_start = None
    if split_tokens[0] != '':
        # There was text before the first escape sequence, this portion does
        # not need to be coloured
        saved_start = split_tokens[0]
        split_tokens = split_tokens[1:]

    try:
        ansi_content_tuples = [re.match(r'(.*?)m(.*)', _s).groups()
                               for _s in split_tokens if _s]
    except AttributeError:
        # No ansi codes in this line. Just return a list of a single tuple,
        # that describes the whole line, with the default cureses attribute
        return [(highlighted_line, curses.A_NORMAL)]

    logging.debug('Tokenized line with ansi codes: %s', ansi_content_tuples)

    join_mergable_ansi_codes(ansi_content_tuples)

    logging.debug('Tokenized line with ansi codes AFTER merging: %s',
                  ansi_content_tuples)

    res = []

    # TODO: Time this code, if it's slow then parallelize it
    # Convert the ansi code portion of each tuple into curses attributes
    for ansi_codes, token_str in ansi_content_tuples:
        logging.debug('Working on ansi coded tuple: (%s, %s)',
                      ansi_codes, token_str)
        if ';' in ansi_codes:
            # There are multiple ansi codes for this string
            ansi_code_list = ansi_codes.split(';')
        else:
            ansi_code_list = [ansi_codes]

        # Up until now the codes were read in from strings, convert them to int
        # str --> int ansi codes
        ansi_code_list = map(int, ansi_code_list)

        # ansi --> curses
        # Sort the ansi codes into fg, bg and modifier. At the moment we don't
        # use the background, but may do so in the future.
        fg_colour, _, modifier = ansi_to_curses(ansi_code_list)

        curses_attrs = curses.A_NORMAL
        if fg_colour:
            curses_attrs = fg_colour
            if modifier:
                logging.debug('Applying modifier: %s', modifier)
                curses_attrs = curses_attrs | modifier

        res.append((token_str, curses_attrs))

    # Add the un-coloured portion of the line we saved earlier if any
    if saved_start:
        res = [(saved_start, curses.A_NORMAL)] + res

    return res


def ansi_to_curses(ansi_code_list):
    """
    Takes a list of ansi codes and split them into the foreground, backgroud
    and modifer (bold, underline, etc) components. Fetch the curses colour
    pair for the fg and bg components (applying bold if it is a "bright"
    version of the colour).
    """
    fg_colour = None
    bg_colour = None
    modifier = None
    for ansi_code in ansi_code_list:
        try:
            curses_attr = ANSI_TO_CURSES[ansi_code]
        except KeyError:
            logging.exception('Got an unknown ansi code from pygments')
            # We ran into an unexpected ansi code, just skip this one
            continue

        if ansi_code in ANSI_FOREGROUND:
            if ansi_code in ANSI_BRIGHT:
                fg_colour = curses.color_pair(curses_attr) | curses.A_BOLD
            else:
                fg_colour = curses.color_pair(curses_attr)
            logging.debug('Detected foreground ansi code: %s', ansi_code)
        elif ansi_code in ANSI_BACKGROUND:
            if ansi_code in ANSI_BRIGHT:
                bg_colour = curses.color_pair(curses_attr) | curses.A_BOLD
            else:
                bg_colour = curses_attr
            logging.debug('Detected background ansi code: %s', ansi_code)
        elif ansi_code in ANSI_MODIFIERS:
            modifier = curses_attr
            logging.debug('Detected modifier ansi code: %s', ansi_code)

    return fg_colour, bg_colour, modifier


def join_mergable_ansi_codes(ansi_content_tuples):
    """
    Sometimes we receive multiple ansi codes intended for a single piece of
    text that are in their own separate escape sequences, instead of being
    bundled together with semicolons.

    Example ansi escaped string:
    # noqa: E501
    [34mfromESC[39;49;00m ESC[04mESC[36m.bblameESC[39;49;00m ESC[34mimportESC[39;49;00m main

    After tokenizing, yields:
    [('34', 'from'), ('39;49;00', ' '), ('04', ''), ('36', '.bblame'), ('39;49;00', ' '), ('34', 'import'), ('39;49;00', ' main')]
                                        ^---------------------------^
                                                  ^^example^^

    You can see ansi code 04 (for underline) was in it's own escaped sequence
    with an empty string next to 36 (blue). Both are meant to apply to the text
    '.bblame', so in the loop below, join cases like this.
    """
    for index, (ansi_code, content) in enumerate(ansi_content_tuples):
        if not content:
            # Empty content, this code likely applies to the next item
            next_index = index + 1
            try:
                ansi_content_tuples[next_index]
            except IndexError:
                # Found an empty content tuple at the end of a line, no need to
                # handle this case.
                continue
            tuple_to_update = ansi_content_tuples[next_index]
            # join the ansi codes together with a semicolon since this is
            # common practice in ansi and the code below already knows how to
            # split it out and handle them.
            ansi_content_tuples[next_index] = (';'.join([ansi_code,
                                                         tuple_to_update[0]]),
                                               tuple_to_update[1])



"""Module that holds the mapping of key press to action.
Some actions are simple and the code needed runs here. Other actions involve
complex operations usually done in the screen object, and are just triggered by
actions.
The ActionTable class knows enough about each action (it's name, key, a short
description, the modes it applies to, etc) such that the help screen can be
generated entirely programatically."""

import curses
from curses import ascii as curses_ascii
import types
import logging

from . import modes
from .content import BaseContent
from .options import (UserOptionManager, SYNTAX_HIGHLIGHT_OPT,
                      SHORTEN_FILEPATHS_OPT)


CURSES_KEY_TO_STR = {key_int: key_str[4:] for key_str, key_int in
                     list(vars(curses).items())
                     if key_str.startswith('KEY_')}
# Add mapping for escape
CURSES_KEY_TO_STR[27] = 'ESC'


# I don't love this class here, but seems silly to create a module for it and
# since the help screen is generated from Action class, there is already a fair
# amount of help code here.
class Help(BaseContent):
    """Class to represent help content.
    Basically just the base class"""
    def __init__(self, help_lines):
        self.lines = self.buildlinesdict(help_lines)
        # Run any setup from the super class
        super(Help, self).__init__()

    def buildlinesdict(self, help_lines):
        """No interesting processing/manipulation done to the lines of the
        help screen before being displayed"""
        return super(Help, self).buildlinesdict(help_lines)


class Action(object):
    """Object to represent a action the user can take in the bblame
    application"""
    def __init__(self, name, func, keys,
                 # NOTE: No default value to force the developer to think
                 # about which modes an action should be usable from, as this
                 # has led to a lot of bugs in the past.
                 action_modes,
                 desc='', show_in_help_msg=True):
        """Given inputs, validate they are at least of the right type and
        create a action table entry
        Attributes:
         name': name of action
         desc: short description of action, NOTE: this is shown in help page
         func: associated function to call when user enacts this action
         action_modes: The modes in which this action is functional
         show_in_help_msg: Toggle whether or not this action is in help page
        """
        # If no description was passed use the doc of the function
        if not desc:
            desc = func.__doc__
        assert isinstance(name, str)
        assert isinstance(desc, str)
        assert isinstance(func, types.MethodType)
        assert isinstance(action_modes, list)
        assert isinstance(keys, list)
        assert isinstance(show_in_help_msg, bool)

        self.name = name
        self.desc = desc
        self.func = func
        self.modes = action_modes
        self.show_in_help_msg = show_in_help_msg
        self.keys = keys

    def int_keys(self):
        """Return a list of this actions keys in integer/ascii form"""
        int_keys = []

        for key in self.keys:
            if isinstance(key, int):
                int_keys.append(key)
            elif isinstance(key, str):
                # convert char to int key code
                assert len(key) == 1, "Keys must be a single char!"
                int_keys.append(ord(key))
            else:
                raise Exception('Unexpected key type')

        return int_keys

    def str_keys(self):
        """Return a list of this actions keys in str form"""
        str_keys = []

        for key in self.keys:
            if isinstance(key, str):
                str_keys.append(key)
            elif isinstance(key, int):
                str_key = CURSES_KEY_TO_STR.get(key)
                if str_key:
                    str_keys.append(str_key)
            else:
                raise Exception('Unexpected key type')

        return str_keys


class ActionTable(object):
    """a class for accepting keys and performing actions"""
    def __init__(self, app_stdscr, scrobj):
        self.screen = scrobj
        self.app_stdscr = app_stdscr
        self._init_action_tables()

    # Not using self here, but I don't want this to be a staticmethod, since
    # that fails the method instance check inside Action
    # pylint: disable=no-self-use
    def quit_action(self):
        """Quit the application"""
        # For now just exit
        exit(0)

    def resize_action(self):
        """Resize the screens on a resize event"""
        logging.info('RESIZE EVENT')
        self.screen.resize_windows(self.app_stdscr)
        self.screen.redraw_screen()

    def move_up_action(self):
        """move the screen or visual select cursor up"""
        # if mode is visual, move line highlight
        if self._if_mode_visual():
            self.screen.move_cursor_up()
        # if mode is normal, move screen
        else:
            self.screen.move_scr_up()
        return True

    def move_down_action(self):
        """move the screen or visual select cursor down"""
        # if mode is visual, move line highlight
        if self._if_mode_visual():
            self.screen.move_cursor_down()
        # if mode is normal, move screen
        else:
            self.screen.move_scr_down()
        return True

    def help_action(self):
        """Display the help message"""
        if self._if_mode_help():
            self.screen.status_bar_toast('Already displaying help!')
        else:
            help_msg = self.generate_help()
            self.screen.display_help(Help(help_msg))

    def filepath_action(self):
        """Toggle filepath on or off. Showing long file paths or just file
        names"""
        opt_mngr = UserOptionManager()

        # Toggle highlighting on or off
        filepath_toggle_opt = opt_mngr.get_option(SHORTEN_FILEPATHS_OPT)
        filepath_toggle_opt.toggle_option()
        logging.info('Filepath toggle: %s', filepath_toggle_opt.is_enabled())

    def syntax_action(self):
        """Toggle syntax highlighting on or off. Showing or hiding syntax
        highlighting ONLY IF IT IS ENABLED."""
        screen = self.screen
        opt_mngr = UserOptionManager()
        if not screen.syntax_highlighting_enabled:
            # Syntax highlighting is disabled entirely by user from cli
            screen.status_bar_toast('Syntax highlighting is disabled, cannot '
                                    'toggle')
            return

        # Toggle highlighting on or off
        syntax_toggle_opt = opt_mngr.get_option(SYNTAX_HIGHLIGHT_OPT)
        syntax_toggle_opt.toggle_option()
        logging.info('Syntax toggle: %s', syntax_toggle_opt.is_enabled())

        # Toast to the user the syntax lexer we're using for highlighting
        if screen.mode in [modes.MODE_NORMAL, modes.MODE_VISUAL]:
            lexer = screen.get_current_scr_content().lexer
            if not lexer:
                # Syntax highlighting is enabled but we couldn't determine a
                # lexer to use, either the file isn't formatted correctly,
                # it's just plain text, or pygments doesn't have a lexer for
                # this file type. Let the user know this, so they have some
                # explanation for why nothing happened after they ran the
                # syntax toggle action.
                screen.status_bar_toast('Could not determine file syntax, not '
                                        'highlighting')
                return
            else:
                # If toggling on syntax highlighting, toast to the user which
                # syntax lexer we're using
                if syntax_toggle_opt.is_enabled():
                    screen.status_bar_toast('Highlighting with %s syntax'
                                            % lexer.name)

    def escape_action(self):
        """Return to Normal mode"""
        if self.screen.mode in [modes.MODE_VISUAL]:
            self.screen.mode = modes.MODE_NORMAL
        elif self.screen.mode in [modes.MODE_SHOW, modes.MODE_HELP]:
            self.screen.restore_prev_content_and_state()

    def _if_mode_normal(self):
        """check if the screen obj is in NORMAL mode"""
        return self.screen.mode == modes.MODE_NORMAL

    def _if_mode_visual(self):
        """check if the screen obj is in VISUAL mode"""
        return self.screen.mode == modes.MODE_VISUAL

    def _if_mode_show(self):
        """check if the screen obj is in SHOW mode"""
        return self.screen.mode == modes.MODE_SHOW

    def _if_mode_help(self):
        """check if the screen obj is in HELP mode"""
        return self.screen.mode == modes.MODE_HELP

    def action_mode_warning(self, action):
        """Log and warn the user that the action they are trying to perform
        isn't available in the current mode they're in"""
        log_msg = ('Action "%s" not available in %s mode!'
                   % (action.name, self.screen.mode))
        logging.info(log_msg)
        self.screen.status_bar_toast(log_msg)

    def process_key(self, key):
        """Take a key and check if we have a corresponding action, if so call
        the action function, and log the keypress"""
        isascii = curses_ascii.isascii(key)
        ascii_key = chr(key) if isascii else None
        logging.info('KEY PRESSED (int, ascii), (%s, %s)', key, ascii_key)
        action = self.key_to_action.get(key, None)
        if action:
            logging.info('Associated action: %s', action.name)
            if self.screen.mode not in action.modes:
                self.action_mode_warning(action)
            else:
                logging.info('Calling action func: %s', action.name)
                return action.func()

        logging.info('Key not in action table')

    def generate_help(self):
        """Automagically generate a help document from the action descriptions
        in the action table"""
        help_lines = []
        help_lines.append('KEYS: ACTION - DESCRIPTION')
        help_lines.append('--------------------------')

        for action in self.actions:
            if action.show_in_help_msg:
                keys_str = ', '.join([_f for _f in action.str_keys() if _f])
                help_lines.append(' %s:   %s' % (keys_str, action.name))
                for desc_line in action.desc.splitlines():
                    help_lines.append('    %s' % (desc_line.strip()))
                help_lines.append('')

        return help_lines

    def _init_action_tables(self):
        """From a list of Action objects generate a dict that maps
        single keys to the action that they perform"""
        self.actions = [
            Action('Quit', self.quit_action,
                   ['q'], modes.ALL_MODES),
            Action('Screen Resize', self.resize_action,
                   [curses.KEY_RESIZE], modes.ALL_MODES,
                   show_in_help_msg=False),
            Action('Search', self.screen.get_search_string,
                   ['/'], modes.ALL_MODES_BUT_HELP,
                   'Search downward through the current blame or commit'),
            Action('Next Search Match', self.screen.jump_to_next_search_match,
                   ['n'], modes.ALL_MODES_BUT_HELP,
                   'Jump to the next search match (in the downward '
                   'direction)'),
            Action('Prev Search Match', self.screen.jump_to_prev_search_match,
                   ['N'], modes.ALL_MODES_BUT_HELP,
                   'Jump to the prev search match (in the upward direction)'),
            Action('Visual Select Mode', self.screen.init_vis_cursor,
                   ['v'], [modes.MODE_NORMAL],
                   'Enter visual select mode (only from normal mode)'),
            Action('Show/View Commit', self.screen.init_git_show,
                   ['o'], [modes.MODE_VISUAL],
                   'Show a commit selected by the visual mode cursor'),
            Action('Show/View file Commit', self.screen.init_git_show_file,
                   ['O'], modes.NORMAL_AND_VISUAL,
                   'Show the current revision commit'),
            Action('Normal Mode', self.escape_action,
                   [27], modes.ALL_MODES),
            Action('Drill Down', self.screen.add_blame_drill,
                   [10, curses.KEY_ENTER, 'd'], [modes.MODE_VISUAL],
                   'Drill down past the commit highlighted in visual mode. '
                   'Opens a new git blame'),
            Action('Parent blame', self.screen.add_blame_parent,
                   ['<', ','], modes.NORMAL_AND_VISUAL,
                   'Move to git blame of the parent of current commit \n'
                   '(i.e. traverse backwards through history, one commit at '
                   'a time)'),
            Action('Ancestor blame', self.screen.add_blame_ancestor,
                   ['>', '.'], modes.NORMAL_AND_VISUAL,
                   'Move to git blame of the ancestor of current commit \n'
                   '(i.e. traverse forwards through history, one commit at '
                   'a time)'),
            Action('Pop Back', self.screen.restore_prev_content_and_state,
                   [8, curses.KEY_BACKSPACE, curses.KEY_DC, 127, 'f'],
                   modes.ALL_MODES,
                   'Pop back to previous git object'),
            Action('Move Up', self.move_up_action,
                   ['k', curses.KEY_UP], modes.ALL_MODES,
                   show_in_help_msg=False),
            Action('Move Down', self.move_down_action,
                   ['j', curses.KEY_DOWN], modes.ALL_MODES,
                   show_in_help_msg=False),
            Action('Move Up Page', self.screen.move_scr_up_page,
                   [curses.KEY_PPAGE], modes.ALL_MODES,
                   'Move the screen up one page',
                   show_in_help_msg=False),
            Action('Move Down Page', self.screen.move_scr_down_page,
                   [curses.KEY_NPAGE], modes.ALL_MODES,
                   'Move the screen down one page',
                   show_in_help_msg=False),
            Action('Move Right Page', self.screen.move_scr_right_nchars,
                   [curses.KEY_RIGHT], modes.ALL_MODES,
                   'Move the screen half a page to the right',
                   show_in_help_msg=False),
            Action('Move Left Page', self.screen.move_scr_left_nchars,
                   [curses.KEY_LEFT], modes.ALL_MODES,
                   'Move the screen half a page to the left',
                   show_in_help_msg=False),
            Action('Jump to Top', self.screen.move_scr_to_top,
                   ['g', curses.KEY_HOME], modes.ALL_MODES,
                   'Jump to the top of the screen'),
            Action('Jump to Bottom', self.screen.move_scr_to_bottom,
                   ['G', curses.KEY_END], modes.ALL_MODES,
                   'Jump to the bottom of the screen'),
            Action('Help', self.help_action,
                   ['h'], modes.ALL_MODES),
            Action('Toggle Syntax Highlight', self.syntax_action,
                   ['s'], modes.ALL_MODES),
            Action('Toggle filepaths/filenames', self.filepath_action,
                   ['b'], modes.NORMAL_AND_VISUAL),
            Action('Jump to HEAD', self.screen.add_head_blame,
                   ['H'], modes.NORMAL_AND_VISUAL,
                   'Jump to a blame of the most recent commit for the file'),
            Action('Jump to TAIL', self.screen.add_tail_blame,
                   ['T'], modes.NORMAL_AND_VISUAL,
                   'Jump to a blame of the first commit for the file'),
        ]

        # Create a mapping of each single key (as its int keycode) to
        # the action, to make single key lookups easier in main loop
        self.key_to_action = {}
        for action in self.actions:
            for key in action.int_keys():
                if key in self.key_to_action:
                    raise Exception('Duplicate action key! %s mapped to %s '
                                    'and %s' % (key, action.name,
                                                self.key_to_action[key].name))
                self.key_to_action[key] = action



"""A module containing functions related to search"""
import os
import curses
import logging

from . import util

SEARCH_STRING_JUMP_FORWARD = 'forward'
SEARCH_STRING_JUMP_BACKWARD = 'backward'
BBLAMEHST_PATH = os.path.join(os.path.expanduser('~'), '.bblamehst')
HISTORY_LIMIT = 50


def find_search_locs(scr_content, search_str):
    """find all the indexes of lines that contain the search string"""
    search_locs = []
    if search_str:
        for idx, line in enumerate(scr_content):
            if search_str in line.full_text():
                search_locs.append(idx)
            if idx == len(scr_content) - 1:
                break
    return search_locs


def re_create_hist_dir():
    """Unlinks the current hist dir if any, and re-create it.
    Called when something has gone wrong with the history file, either someone
    has changed the file permissions, ownership, etc. Just start fresh with a
    new file."""
    try:
        os.unlink(BBLAMEHST_PATH)
    except OSError as err:
        # If the  error is anything but file doesn't exist, then we have an
        # issue.
        if err.errno != 2:
            raise err
    # Essentially just "touch" the file to create it
    open(BBLAMEHST_PATH, 'a').close()


def read_search_history():
    """Load the stored search history from disk. File is .bblamehst.
    History is limited to 50 entries"""
    try:
        with open(BBLAMEHST_PATH, 'r') as bblamehst_file:
            return bblamehst_file.read().splitlines()
    # pylint: disable=bare-except
    except:  # noqa: E722
        re_create_hist_dir()
        return []


def write_search_history(history):
    """Write the stored search history to disk. File is ~/.bblamehst
    History is limited to 50 entries"""
    def write_to_hist_file():
        with open(BBLAMEHST_PATH, 'w') as bblamehst_file:
            for search_pattern in history:
                bblamehst_file.write(search_pattern + '\n')
    try:
        write_to_hist_file()
    # pylint: disable=bare-except
    except:  # noqa: E722
        re_create_hist_dir()
        try:
            write_to_hist_file()
        # pylint: disable=bare-except
        except:  # noqa: E722
            # Give up at this point, but don't kill the whole application.
            # Since this isn't a crucial feature.
            pass


def add_search_str_to_history(search_history, search_str):
    """Add the search_str to the search history to be written to disk.
    Ensuring to keep the history <= the history limit"""
    # search_history.append(search_str)
    search_history.insert(0, search_str)
    logging.info('search history: %s', search_history)
    return search_history[:HISTORY_LIMIT]


def append_string(txtbox, str_to_append):
    """Add string to the texteditpad textbox. This is not an append, will
    overwrite the existing text"""
    txtbox.text = [str_to_append]
    set_txtbox_cursor_pos(txtbox, 0, len(str_to_append))
    txtbox.redraw_vlines(txtbox.vptl, (0, 0))


def clear_search_txtbox(txtbox):
    """Clear the current text from the txtbox"""
    if len(''.join(txtbox.text)) > 0:
        txtbox.text = ['']
        set_txtbox_cursor_pos(txtbox, 0, 0)
        txtbox.redraw_vlines(txtbox.vptl, (0, 0))


def set_txtbox_cursor_pos(txtbox, x_pos, y_pos):
    """Set the cursor position of the texteditpad cursor"""
    txtbox.ppos = (x_pos, y_pos)  # physical position of the cursor
    txtbox.vpos = (x_pos, y_pos)  # virtual position of the cursor


def collect_search_string(txtbox):
    """Collect the search string from the user. Handling each character
    inputted into the txtbox, and handling the result
    """
    search_history = read_search_history()
    logging.info('SEARCH: history list: %s', search_history)
    history_iter = util.BidirectionalCycle(search_history, no_wrap=True)
    logging.info("SEARCH: history iter %s", history_iter)
    while True:
        char = txtbox.win.getch()
        if len(''.join(txtbox.text)) == 0:
            history_enabled = True
        # Enter key breaks out of the collection loop
        if char == curses.KEY_ENTER or char == ord('\n'):
            break
        # Handle backspace here if the textbox is empty, if so, exit the
        # collection loop
        elif char in [8, curses.KEY_BACKSPACE, curses.KEY_DC] and \
                len(''.join(txtbox.text)) == 0:
            return
        # Handle key up and down here if the textbox is empty, if so, cycle
        # through the past search history
        elif char in [curses.KEY_DOWN,
                      curses.KEY_UP] and history_iter and history_enabled:
            # clear_search_txtbox(txtbox)
            try:
                if char == curses.KEY_UP:
                    next_search = next(history_iter)
                else:
                    next_search = history_iter.prev()
            except StopIteration:
                # We've hit the end of the history, just return the current
                # item instead of advancing to the next or previous
                next_search = history_iter.curr()
            append_string(txtbox, next_search)
            logging.info('SEARCH: history next: %s', next_search)
        # Pass the key off to the textpad to handle from here
        else:
            history_enabled = False
            txtbox.do_command(char)

    search_str = ''.join(txtbox.text)
    # It's not helpful to add adjacent duplicates to search history, so only
    # add an item if it's unique compared to the last added item
    if not search_history or search_history[0] != search_str:
        search_history = add_search_str_to_history(search_history, search_str)
    write_search_history(search_history)

    return search_str



"""A main file for executing bblame from within the package without deploying
bblame as a binary or with distutils.
e.g. from root of the project dir: python -m betterblame"""
from .bblame import main
main()



"""Classes and utility functions related to git data and operations on it"""
import os
import re
import curses
import logging
from datetime import datetime
from collections import namedtuple


from .highlight import get_lexer, highlight_lines
from .content import Line, BaseContent
from .util import left_str_strip
from .sh import sh


class BadRevException(Exception):
    """Simple custom exception to make exception handling easier up the
    stack"""
    pass


class NoSuchRevException(Exception):
    """Simple custom exception to make exception handling easier up the
    stack"""
    pass


NUM_TAB_SPACES = 4
ABBREV_LEN = 8
# Git blame appears to show always n+1 characters of the sha, where n is
# the number of chars you actually asked for...
ABBREV_LEN_BLAME = ABBREV_LEN - 1
# Git color pair names
GIT_COLOR_OLD = curses.COLOR_RED
GIT_COLOR_NEW = curses.COLOR_GREEN
GIT_COLOR_FRAG = curses.COLOR_CYAN
GIT_COLOR_FUNC = curses.COLOR_BLUE
GIT_COLOR_SHOW = curses.COLOR_YELLOW
GIT_COLOR_META = curses.COLOR_MAGENTA


def get_head_rev(filename):
    """Get the revision pointed to by HEAD for <filename> specifically"""
    cmd = ['rev-list', '--abbrev=%d' % ABBREV_LEN, '--max-count=1',
           'HEAD', '--', filename]
    git = sh.git.bake(_cwd=os.getcwd())
    revlist = git(*cmd)
    if not revlist.splitlines():
        raise NoSuchRevException('No revlist for HEAD for file: %s \n'
                                 % filename)
    else:
        # return the firs 8 chars of the revision
        return revlist.splitlines()[-1][:8]


class GitLog(object):
    """A class to represent the git log.
    Essentially a hash mapped linked list. A git sha is the key to a gitlog
    object, which can then traverse to it's parent or ancestor, and so on"""

    class LogEntry(object):
        """A class object to represent an entry of the git log.
        Stores the sha and filename for that commit, and references to this
        commits parent and ancestor"""
        def __init__(self, sha, filename, desc, parent=None, ancestor=None):
            self.sha = sha
            self.filename = filename
            self.desc = desc
            self.parent = parent
            self.ancestor = ancestor

    def __init__(self, initial_filename):
        self._gitlog = {}
        git = sh.git.bake('--no-pager', _cwd=os.getcwd())
        initial_log = git.log('--no-color', '--abbrev=%d' % ABBREV_LEN,
                              '--follow', '--oneline', '--name-only', '--',
                              '%s' % initial_filename)
        logging.info('Preparing Git Log')
        previous_log_entry = None
        curr_log_entry = None
        # Fancy syntax to iterate two lines at a time
        for sha_and_desc, filename in zip(*[iter(initial_log.splitlines())]*2):
            logging.info('sha_and_desc: %s', sha_and_desc)
            sha = sha_and_desc.split()[0]
            desc = ' '.join(sha_and_desc.split()[1:])
            curr_log_entry = self.LogEntry(sha, filename, desc,
                                           ancestor=previous_log_entry)
            if previous_log_entry:
                previous_log_entry.parent = curr_log_entry
            else:
                # Keep a reference to the head and tail (see below). To be used
                # for the action which snaps bblame to the first or most recent
                # commit for a file
                self.head = curr_log_entry
            previous_log_entry = curr_log_entry

            self._gitlog[sha] = curr_log_entry

        if curr_log_entry:
            self.tail = curr_log_entry

    def __getitem__(self, item):
        return self._gitlog[item]

    def __str__(self):
        return self._gitlog.__str__()

    def __iter__(self):
        return self._gitlog.__iter__()

    def __len__(self):
        return self._gitlog.__len__()


PorcelainChunk = namedtuple('PorcelainChunk',
                            ['author', 'author_mail', 'author_tz',
                             'author_time', 'sha', 'content', 'filepath',
                             'filename', 'orig_line_no', 'final_line_no'])


LineMetaData = namedtuple('LineMetaData',
                          ['sha', 'filename', 'orig_line_no'])


class Blame(BaseContent):
    """A class to represent a blame.

    Attributes:
        git_sha - The SHA of the git commit being blamed if one
        filename - The name of the file we are blaming
        lines - The output of git blame by line in a default dict
        lexer - The syntax lexer used to highlight this blame
        lines_metadata - Extra info for each line (e.g. sha, filename, etc)
        max_filename_len - used to pad the blame output
        max_author_len - used to pad the blame output
        syntax_highlighting_enabled - whether syntax highlighting is enabled
                                      across the app
    """

    def __init__(self, filename, syntax_highlighting_enabled, git_sha=''):
        git = sh.git.bake('--no-pager', _cwd=os.getcwd())
        self.filename = filename
        self.lines_metadata = {}
        self.max_filename_len = 0
        self.max_filepath_len = 0
        self.max_author_len = 0
        self.syntax_highlighting_enabled = syntax_highlighting_enabled

        if git_sha:
            try:
                cmd = [git_sha, '--line-porcelain',
                       '--abbrev=%d' % ABBREV_LEN_BLAME, '--', filename]
                self.lines = self.buildlinesdict(
                    git.blame(*cmd).stdout.splitlines())
            except sh.ErrorReturnCode_128 as exc:
                stderr = exc.stderr.decode('utf-8')
                if "no such path" in stderr:
                    raise NoSuchRevException(stderr)
                if "bad revision" in stderr:
                    raise BadRevException(stderr)
                raise
            self.git_sha = git_sha
        else:
            self.git_sha = get_head_rev(filename)
            self.lines = self.buildlinesdict(git.blame(
                '--line-porcelain', '--abbrev=%d' % ABBREV_LEN_BLAME,
                filename).stdout.splitlines())

        # Run any setup from the super class
        super(Blame, self).__init__()

        # Clean up git we don't need it anymore
        del git

    def process_porcelain_blame(self, lines):
        """Process raw line-porcelain output into porcelain chunk objects"""
        ret_list = []
        start_idx = 0

        logging.info('Processing porcelain blame')

        for idx, line in enumerate(lines):
            line = line.decode('utf-8')
            logging.info('Looking at line: %s', line)
            if re.match(r'\t.*', line):
                logging.info('FOUND A TAB')
                # Line starts with tab, we've hit the end of a porcelain chunk
                chunk_lines = lines[start_idx:idx+1]
                ret_list.append(self.process_porcelain_chunk(chunk_lines))
                logging.info('porcelain lines: %s', lines[start_idx:idx])
                start_idx = idx + 1
                logging.info(ret_list[-1])

        logging.info('ret_list: %r', ret_list)
        return ret_list

    def process_porcelain_chunk(self, chunk_lines):
        """Process a single porcelain chunk into a named tuple"""
        header_line_tokens = chunk_lines[0].decode('UTF-8').split()
        if len(header_line_tokens) == 3:
            sha, orig_line_no, final_line_no = header_line_tokens
        elif len(header_line_tokens) == 4:
            (sha, orig_line_no,
             final_line_no, mystery_value) = header_line_tokens
        else:
            raise Exception('Header of porcelain chunk has unexpected format!')

        content = chunk_lines[-1][1:].decode('UTF-8')
        # Convert tabs to spaces, tabs were unpredictably expanding and
        # causing issues
        content = content.replace('\t', ' '*NUM_TAB_SPACES)
        metadata = chunk_lines[:-1]
        author = None
        author_mail = None
        author_tz = None
        author_time = None
        filepath = None

        # Simple processing since the format is known
        for line in metadata:
            logging.info('processing chunk line: %s', line)
            line = line.decode('UTF-8')

            if not author and line.startswith('author '):
                author = left_str_strip(line, 'author ')
                logging.info('author: %s', author)
                continue

            if not author_tz and line.startswith('author-tz '):
                author_tz = left_str_strip(line, 'author-tz ')
                logging.info('author_tz: %s', author_tz)
                continue

            if not author_mail and line.startswith('author-mail '):
                author_mail = left_str_strip(line, 'author-mail ')
                logging.info('author_mail: %s', author_mail)
                continue

            if not author_time and line.startswith('author-time '):
                author_time = left_str_strip(line, 'author-time ')
                logging.info('author_time: %s', author_time)
                continue

            # Git calls this a filename but it is really a path, so capture
            # both here
            if not filepath and line.startswith('filename '):
                filepath = left_str_strip(line, 'filename ')
                filename = os.path.basename(filepath)
                continue

        if not all([author, author_tz, author_time, author_mail,
                    filepath, orig_line_no, final_line_no]):
            logging.debug('author: %s, author_tz: %s, author_time: %s, '
                          'author_mail: %s, content: %s, filepath: %s, '
                          'orig_line_no: %s, final_line_no: %s',
                          author, author_tz, author_time, author_mail,
                          content, filepath, orig_line_no,
                          final_line_no)
            raise Exception('Couldn\'t extract all needed data from git blame')

        ret = PorcelainChunk(author=author, author_mail=author_mail,
                             content=content, author_tz=author_tz,
                             author_time=author_time, sha=sha,
                             orig_line_no=orig_line_no,
                             final_line_no=final_line_no,
                             filepath=filepath, filename=filename)

        if len(author) > self.max_author_len:
            self.max_author_len = len(author)
        if len(filepath) > self.max_filepath_len:
            self.max_filepath_len = len(filepath)
        if len(filename) > self.max_filename_len:
            self.max_filename_len = len(filename)

        return ret

    def buildlinesdict(self, lines):
        """Construct the dictionary that holds the lines for the Blame. See
        documentation in BaseContent for more details"""
        ret = self.gen_default_lines_dict()
        # The filename is not included in the format string below since
        # it can be shortened by the user at run time. So it is a separate
        # Segment and tagged so it can be identified later when drawing
        metadata_segment_txt = '(%s %s %s %s) '
        porcelain_chunks = self.process_porcelain_blame(lines)
        num_lines = len(porcelain_chunks)

        if self.syntax_highlighting_enabled:
            content_lines = [chunk.content for chunk in porcelain_chunks]
            logging.info('Content lines for blame before highlighting: %s',
                         content_lines)

            self.lexer = get_lexer(content_lines, self.filename)
            if self.lexer is None or 'text/plain' in self.lexer.mimetypes:
                # Can't determine a lexer, so just use normal highlight
                self.lexer = None
                highlighted_lines = [[(line, curses.A_NORMAL)]
                                     for line in content_lines]
            else:
                logging.info('Using %s lexer', self.lexer.name)
                highlighted_lines = highlight_lines(content_lines, self.lexer)
                logging.info('Highlighted lines for blame: %s',
                             highlighted_lines)

        for idx, porcelain_chunk in enumerate(porcelain_chunks):
            logging.info('Building line chunk %s, at idx: %s',
                         porcelain_chunk.content, idx)

            _sha = porcelain_chunk.sha[:ABBREV_LEN_BLAME+1]
            _author = porcelain_chunk.author.ljust(self.max_author_len)
            _filename = porcelain_chunk.filename.ljust(self.max_filename_len)
            _filepath = porcelain_chunk.filepath.ljust(self.max_filepath_len)
            _line_no = porcelain_chunk.final_line_no.rjust(len(str(num_lines)))
            dtime = datetime.fromtimestamp(float(porcelain_chunk.author_time))
            _author_time = dtime.strftime('%Y-%m-%d %H:%M:%S')

            line = Line()
            line.add_segment(_sha, curses.A_DIM)
            line.add_filepath_segment(filepath_text=' %s ' % _filepath,
                                      filename_text=' %s ' % _filename,
                                      segment_attributes=curses.A_DIM)
            line.add_segment(metadata_segment_txt % (_author,
                                                     _author_time,
                                                     porcelain_chunk.author_tz,
                                                     _line_no), curses.A_DIM)

            if self.syntax_highlighting_enabled:
                highlighted_line = highlighted_lines[idx]
                logging.info('highlighted content: %s', highlighted_line)
                for string, syntax_attrs in highlighted_line:
                    line.add_segment(string, syntax_attrs=syntax_attrs)
            else:
                line.add_segment(porcelain_chunk.content)

            logging.info('bline segments:: %s', line.line_segments)
            ret[idx] = line

            # keep track of some of the most important metadata for each line
            line_meta = LineMetaData(sha=_sha,
                                     filename=porcelain_chunk.filename,
                                     orig_line_no=porcelain_chunk.orig_line_no)
            self.lines_metadata[idx] = line_meta

        return ret


class Show(BaseContent):
    """A class to represent a git show.

    Attributes:
        showobj - The function object returned by sh.git, contains attributes
                   like the return code of the command, the stdout, etc.
        git_sha - The SHA of the git commit being showed
    """
    def __init__(self, git_sha):
        git = sh.git.bake('--no-pager', _cwd=os.getcwd())
        showobj = git.show('--no-color', git_sha)
        self.git_sha = git_sha
        self.lines = self.buildlinesdict(showobj.stdout.splitlines())

        # Run any setup from the super class
        super(Show, self).__init__()

        # Clean up git and showobj we don't need it anymore
        del git
        del showobj

    def buildlinesdict(self, lines):
        """Construct the dictionary that holds the lines for the Show. See
        documentation in BaseContent for more details.
        Colourize the lines appropriately for the different types of Show lines
        """
        ret = self.gen_default_lines_dict()
        for idx, line in enumerate(lines):

            if isinstance(line, bytes):
                line = line.decode('UTF-8')

            # Check for the various git show colours
            # Commit header
            if line.startswith('commit'):
                ret[idx] = Line([line], curses.color_pair(GIT_COLOR_SHOW))

            # Meta information block
            elif line.startswith(('diff', '---', '+++', 'index', 'new file')):
                ret[idx] = Line([line], curses.color_pair(GIT_COLOR_META))

            # hunk/fragment header
            elif line.startswith('@@'):
                header_match = re.match(r'(@@ .* @@)(.*)', line)
                # should have two groups, who together are the same length as
                # the original line (to ensure we captured all the text)
                if (header_match and len(header_match.groups()) == 2
                        and sum(map(len, header_match.groups())) == len(line)):

                    # Use two segments and paint them different colours
                    ret[idx] = Line([header_match.groups()[0]],
                                    curses.color_pair(GIT_COLOR_FRAG))
                    ret[idx].add_segment(header_match.groups()[1],
                                         curses.color_pair(GIT_COLOR_FUNC))

                # We couldn't properly parse the hunk header, just paint the
                # whole line the same colour
                else:
                    ret[idx] = Line([line], curses.color_pair(GIT_COLOR_FRAG))

            # added lines
            elif line.startswith('+'):
                ret[idx] = Line([line], curses.color_pair(GIT_COLOR_NEW))

            # removed lines
            elif line.startswith('-'):
                ret[idx] = Line([line], curses.color_pair(GIT_COLOR_OLD))

            # line with no interesting colours
            else:
                ret[idx] = Line([line])

        return ret



"""Different modes of the bblame application"""
# Normal/standard view
MODE_NORMAL = 'normal'
# Visual select mode for selecting lines of the blame file
MODE_VISUAL = 'visual'
# Viewing a show not a blame
MODE_SHOW = 'show'
# Viewing a help message
MODE_HELP = 'help'

# Useful lists of modes
ALL_MODES = [MODE_NORMAL, MODE_VISUAL, MODE_SHOW, MODE_HELP]
ALL_MODES_BUT_HELP = [MODE_NORMAL, MODE_VISUAL, MODE_SHOW]
NORMAL_AND_VISUAL = [MODE_NORMAL, MODE_VISUAL]



"""A module to contain objects and implementation which together are
the backbone of the internal representation of the text the user sees"""
import curses
import re
import logging
import gc
import zlib

from abc import ABCMeta, abstractmethod
from collections import defaultdict
from collections import namedtuple
from six import string_types

from .options import UserOptionManager, SHORTEN_FILEPATHS_OPT
from .util import substring_split

DEFAULT_CURSES_ATTR = curses.A_NORMAL


class ContentStack(object):
    """A class to represent the content added as users drill into git revs,
    cycle through git history, show commits, etc"""
    def __init__(self):
        self.content_stack = []

    # These frames also store the state necessary to move back to the
    # previous frame/screen when it is popped. Essentially storing a snapshot
    # of what the screen looked like before we add this frame
    StackFrame = namedtuple('StackFrame', ['content', 'last_current_line',
                                           'last_current_width',
                                           'last_cursor_line', 'last_mode'])

    def __len__(self):
        return len(self.content_stack)

    def peek(self):
        """Peek and return the content on the top of the stack,
        without removing it from the stack"""
        return self.content_stack[-1]

    def pop(self):
        """Pop content off the content stack, first checking if we have any
        extra content that can be popped (I.e. we don't want to be left]
        without any content to display to the user."""
        cstack = self.content_stack
        if len(cstack) <= 1:
            raise IndexError('No content to pop')
        frame_to_return = cstack.pop()
        self.content_stack[-1].content.decompress()
        return frame_to_return

    def add(self, newcontent_obj, current_line, current_width,
            cursor_line, mode):
        """Add a new frame object to the Content Stack, preserving some of our
        current state as we do that, so it can be restored when we pop this."""
        # compress current stack before adding the new stack
        if len(self.content_stack) >= 1:
            last_content = self.content_stack[-1].content
            # If the last content was a blame, then compress it as these can
            # be quite large.
            last_content.compress()
        stack_frame = self.StackFrame(newcontent_obj, current_line,
                                      current_width, cursor_line, mode)
        self.content_stack.append(stack_frame)


class BaseContent:
    """A base class for the git objects (blame and show)
    Comes with len and get/set overrides and the methods to compress and
    uncompress the lines attribute"""
    __metaclass__ = ABCMeta

    default_attr = curses.A_NORMAL
    default_str = '~'

    @abstractmethod
    def __init__(self):
        # To be called after subclasses __init__ work is done

        # self.lines comes from subclasses, init it again here to make pylint
        # happy
        self.lines = self.lines  # Satisfy pylint

        # lines will grow if any '~' padding is added
        self.numlines = len(self.lines)
        # Collect after any deleted objects
        gc.collect()

    @abstractmethod
    def buildlinesdict(self, lines):
        """build a default dict with schema:
        index --> Line(line or <defaultstr>, curses str attributes)
        The default string makes scrolling past the edge of the file much more
        graceful.
        I.e. having a default value rather than handling IndexError and then
        returning the default string.
        This function will add a default or the provided attribute to each
        line.  The lines can be further decorated at later times by logically
        OR-ing in other attributes to the attribute set here. Classes that need
        to do more advanced processing of the lines, should override this
        method"""
        logging.info('running base buildlinesdict')
        ret_lines_dict = self.gen_default_lines_dict()
        for idx, line in enumerate(lines):
            ret_lines_dict[idx] = Line([line], BaseContent.default_attr)

        return ret_lines_dict

    def gen_default_lines_dict(self):
        """Return a default dict with the proper schema to be used for a lines
        dictionary"""
        return defaultdict(lambda: (Line([BaseContent.default_str],
                                         BaseContent.default_attr)))

    def __len__(self):
        return self.numlines

    def __getitem__(self, index):
        return self.lines[index]

    def __setitem__(self, index, value):
        self.lines[index] = value

    def compress(self):
        """Compress the lines attribute which stores the content of the lines
        attribute. This makes storing each stack frame cheaper"""
        # XXX: compress/decompress currently broken by new Line Segments
        # feature. It's more difficult to compress the lines (with a good
        # compression ratio) as well as keeping the curses attributes for each
        # segment. Will address this feature at a later date (either fixing or
        # removing this feature)
        # pylint: disable=unreachable
        return

        blame_text = '\n'.join([line.full_text()
                                for line in self.lines.values()])
        compressed_lines = zlib.compress(blame_text.encode('utf-8'))
        del self.lines
        gc.collect()
        self.lines = compressed_lines

    def decompress(self):
        """Decompress the lines attribute which stores the content of the git
        blame."""
        # XXX: compress/decompress currently broken by new Line Segments
        # feature. It's more difficult to compress the lines (with a good
        # compression ratio) as well as keeping the curses attributes for each
        # segment. Will address this feature at a later date (either fixing or
        # removing this feature)
        # pylint: disable=unreachable
        return

        temp_lines = zlib.decompress(self.lines)
        # We don't need any special processing at this point, since it was
        # already done and we're just re-hydrating it from concentrate
        self.lines = BaseContent.buildlinesdict(self, temp_lines.splitlines())
        del temp_lines
        gc.collect()


class Segment(object):
    """A class to represent a single segment of a line. Segments represent a
    portion of the text of a line and have their own curses attributes and
    colours"""
    def __init__(self, text, curses_attrs, syntax_attrs):
        """
        <text>: The text the user will see that this segment represents
        <curses_attrs>: Specific curses attributes for this segment
        <syntax_attrs>: Syntax highlighting attributes for colouring text
        This allows you to have a single line be coloured differently or have
        portions standout or bold, etc"""
        assert type(curses_attrs) is int

        self._text = text
        self.curses_attrs = curses_attrs
        self.syntax_attrs = syntax_attrs
        self.opt_mngr = UserOptionManager()

    @property
    def text(self):
        """Getter for text. Must have a getter in this case, because other
        Segment types inherit from this one, so if it sets text when it is
        being called from a super init, it will cause an attribute error"""
        return self._text


class FilepathSegment(Segment):
    """A segment that represents a file path. Which can return the full path or
    just file name depending on user options"""
    def __init__(self, filepath_text, curses_attrs, syntax_attrs,
                 filename_text):
        # The first arg passed to init will be set as self._text
        super(FilepathSegment, self).__init__(filepath_text, curses_attrs,
                                              syntax_attrs)
        self.filename_text = filename_text

    @property
    def text(self):
        """Getter method for text. Depending on the current state and user
        options enabled this will return different text"""
        shorten_paths_opt = self.opt_mngr.get_option(SHORTEN_FILEPATHS_OPT)
        if shorten_paths_opt.is_enabled():
            return self.filename_text

        return self._text


class Line(object):
    """A class to represent each line of the git object being shown (blame,
    show, etc)
    """

    def __init__(self, list_of_segments=None, default_attr=curses.A_NORMAL):
        """Can provide a list of strings, each string will become one segment
        with its own curses attributes, segments passed this way will get the
        default curses attributes"""
        if not list_of_segments:
            list_of_segments = []
        self.line_segments = []  # a list of line_segment tuples

        # Set the default attribute for this line
        self.default_attr = default_attr

        assert not isinstance(list_of_segments, string_types), \
            'Argument to Line must be a list of strings'

        for segment in list_of_segments:
            if isinstance(segment, Segment):
                # This is already a segment type, so just add it to the list
                self.line_segments.append(segment)
            else:
                # Otherwise init a new segment
                self.add_segment(segment)

    def add_segment(self, segment_text, segment_attributes=None,
                    syntax_attrs=None):
        """Add a new segment to the line. See the Segment class for more
        information"""
        if not segment_attributes:
            segment_attributes = self.default_attr

        self.line_segments.append(Segment(segment_text,
                                          segment_attributes,
                                          syntax_attrs))

    def add_filepath_segment(self, filepath_text, filename_text,
                             segment_attributes=None,
                             syntax_attrs=None):
        """Add a new filepath segment to the line. This type of segment has two
        texts the full path and just the name"""
        if not segment_attributes:
            segment_attributes = self.default_attr

        self.line_segments.append(FilepathSegment(filepath_text,
                                                  segment_attributes,
                                                  syntax_attrs, filename_text))

    def full_text(self):
        """Returns the entire text of the line, combining all text from each
        segments. This returns the line with no default_attr"""
        ret = ''

        for segment in self.line_segments:
            ret = ret + segment.text

        return ret

    def prepare_line_for_printing(self, start, end):
        """Prepare a line that is ready for printing to the screen for the
        user. This involves trimming the line to fit within the <width> of the
        screen, mutating segments or even discarding them if the users wishes)
        """
        num_chars_so_far = 0
        # A throw away line that will be used to print to the screen and
        # discarded
        ret_line = Line()

        if start < 0 or start > end:
            raise Exception('Invalid start arg given')

        for segment in self.line_segments:
            # Mutate the segment or discard it before we use it for length
            # any calculations:
            seg_text = segment.text

            if start > num_chars_so_far:
                # Check if we should begin collecting segments
                if len(seg_text) + num_chars_so_far > start:
                    # Check how much of this segment that just pushes into the
                    # collection range should be included. This is the latter
                    # portion of the segment text
                    chars_to_take = ((len(seg_text) + num_chars_so_far) -
                                     start)
                    ret_line.add_segment(seg_text[-chars_to_take:],
                                         segment.curses_attrs,
                                         segment.syntax_attrs)
            else:
                if len(seg_text) + num_chars_so_far < end:
                    # The whole segment can fit
                    ret_line.add_segment(seg_text,
                                         segment.curses_attrs,
                                         segment.syntax_attrs)
                else:
                    # We can't fit the entirety of the next segment, but check
                    # if a subset will fit. This is the former portion of the
                    # segment.
                    num_chars_remaining = end - num_chars_so_far
                    if num_chars_remaining:
                        ret_line.add_segment(seg_text[:num_chars_remaining],
                                             segment.curses_attrs,
                                             segment.syntax_attrs)
                    break
            num_chars_so_far += len(seg_text)

        return ret_line

    def highlight_str(self, str_to_highlight, highlight_attr):
        """Takes an input string to highlight and searches for instances of it
        (even across segment boundaries), creating a new segment for each hit
        and changing it's curses attribute to highlight.
        Returns a new Line object"""
        # print('string to highlight: %s' % str_to_highlight)
        search_str_re = re.escape(str_to_highlight)
        # print('string to highlight re: %s' % search_str_re)
        segments = self.line_segments
        for start_pos in [match.start() for match in
                          re.finditer('(%s)' % search_str_re,
                                      self.full_text())]:
            # print('working on start_pos: %d' % start_pos)
            segments = self._split_new_segment(start_pos,
                                               start_pos+len(str_to_highlight),
                                               highlight_attr, segments)
        return Line(segments)

    def _split_new_segment(self, start, end, curses_attrs, segments):
        """Cut a new segment out at <start>, <end>, giving it the curses
        attributes <curses_attrs>.
        Returns a new segment list"""
        new_segments = []
        new_seg_len = end-start
        done_split = False
        if start < 0 or start > end:
            raise Exception('Invalid start arg given')

        num_chars_so_far = 0
        remaining_chars = -1
        # print('start %d' % start)
        # print('end %d' % end)
        # print('full_text:  %s' % ':'.join([seg.text for seg in segments]))
        segments = iter(segments)
        for segment in segments:
            # Mutate the segment or discard it before we use it for length
            # calculations
            seg_text = segment.text
            seg_len = len(seg_text)
            # print('working on seg %s, with len: %d' % (segment, seg_len))
            # print('num_chars_so_far: %d' % num_chars_so_far)
            if not done_split and start >= num_chars_so_far:
                # print('checking if we\'ve reached start yet')
                if seg_len + num_chars_so_far > start:
                    # print('we\'ve reached the start!')
                    # We've found the split point
                    (prev_s, new_s,
                     next_s) = substring_split(start-num_chars_so_far,
                                               end-num_chars_so_far,
                                               seg_text)
                    # print('prev_s: %s' % prev_s)
                    if prev_s:
                        prev_seg = Segment(prev_s,
                                           segment.curses_attrs,
                                           segment.syntax_attrs)
                        # print('prev_seg: %s' % prev_seg.text)
                        new_segments.append(prev_seg)

                    # print('new_s: %s' % new_s)
                    # grab the full substring here. We'll skip all segments
                    # that fall into this range down below
                    if not curses_attrs:
                        curses_attrs = segment.curses_attrs
                    new_segments.append(
                        Segment(self.full_text()[start:end],
                                curses_attrs, None))

                    # print('next_s: %s' % next_s)
                    if next_s:
                        next_seg = Segment(next_s,
                                           segment.curses_attrs,
                                           segment.syntax_attrs)
                        # print('next_seg: %s' % next_seg.text)
                        new_segments.append(next_seg)

                    remaining_chars = new_seg_len - len(new_s)
                    done_split = True
                    num_chars_so_far += len(seg_text)
                    continue

            if remaining_chars > 0:
                # print('remaining_chars: %d' % remaining_chars)
                if remaining_chars < seg_len:
                    # print('We need a portion of this segment')
                    # We need a portion of this line
                    # Modify the segment following this one to remove any
                    # pieces that belong to the new segment
                    next_seg = Segment(
                        segment.text[remaining_chars:],
                        segment.curses_attrs,
                        segment.syntax_attrs)
                    new_segments.append(next_seg)
                else:
                    # print('we need all of the chars from this seg, continue'
                    #       ' skipping this line')
                    pass
                remaining_chars -= seg_len
                num_chars_so_far += len(seg_text)
                continue

            new_segments.append(segment)
            num_chars_so_far += len(seg_text)

        return new_segments

    def __len__(self):
        return len(self.line_segments)

    def __getitem__(self, index):
        return self.line_segments[index]

    def __setitem__(self, index, value):
        raise NotImplementedError('__setitem__ not implemented, please use the'
                                  'add_segment method')



"""Entry point/main module that handles initial argument parsing and setup of
the main curses loop, that waits for user input, reacts, and redraws the screen
"""
import curses
import os
import sys
import errno
import argparse
import logging
import traceback
import tempfile

from . import screen
from . import git
from . import actions
from . import version
from .sh import sh

# Setup logging to a temporary dir if we can find one
TEMP_DIR = tempfile.gettempdir()
LOG_FORMAT = ('%(asctime)s|%(levelname)s|'
              '%(module)s.%(funcName)s:%(lineno)d|'
              '%(message)s')
DATE_FORMAT = '%b-%d %H:%M'
LOG_LEVEL = logging.WARNING
if TEMP_DIR:
    logging.basicConfig(filename=os.path.join(TEMP_DIR, 'bblame.log'),
                        format=LOG_FORMAT, datefmt=DATE_FORMAT,
                        level=LOG_LEVEL)
else:
    # Disable all logging
    logging.disable(logging.CRITICAL)


def curses_loop(stdscr, args):
    """main curses application loop to be fed to curses wrapper, which handles
    initializing of the screen and first blame and the user input handling loop
    """
    curses.curs_set(0)
    curses.start_color()
    curses.use_default_colors()
    for i in range(0, curses.COLORS):
        curses.init_pair(i, i, -1)
    scrobj = screen.Screen(stdscr, args.filename, args.revision, args.syntax)
    actiontable = actions.ActionTable(stdscr, scrobj)
    if args.line_num_or_search_term:
        if isinstance(args.line_num_or_search_term, int):
            initial_line = args.line_num_or_search_term
            scrobj.init_line_arg(initial_line)
        else:
            scrobj.search_str = args.line_num_or_search_term
            scrobj.update_search_locs()
            run_and_log_user_errors(scrobj,
                                    scrobj.jump_to_next_search_match)
            scrobj.init_vis_cursor()
    scrobj.redraw_screen()

    # start waiting for and acting on user input
    while True:
        # wait for char from user
        char = scrobj.getch()

        # Process the key with the action table
        skip_redraw = run_and_log_user_errors(scrobj,
                                              actiontable.process_key, char)
        stdscr.erase()
        scrobj.redraw_screen()

        if not skip_redraw:
            # Redrawing on some actions is too slow, the chief example being
            # moving the screen down by one line, if the user is using a mouse
            # wheel to scroll, redrawing the window takes too long, and isn't
            # necessary in that case anyway.
            scrobj.stdscr.redrawwin()

        # log some state each iteration and redraw the screen
        scrobj.log_state()


def run_and_log_user_errors(screenobject, func, *args):
    """Run <func> with <args> in a try/except for UserError, display any such
    errors to the user on the status bar and continue.
    Any other exceptions will crash the application.
    Returns the return value of the function or the UserError that was caught
    and logged."""
    try:
        return func(*args)
    except screen.UserError as exc:
        screenobject.set_status_bar_next_msg(str(exc),
                                             attributes=curses.A_BOLD |
                                             curses.A_STANDOUT)


def process_args():
    """Process the input to bblame, then return the file name to be
    passed to the main function"""

    argparser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
    )
    argparser.add_argument('--revision', '-r',
                           metavar='{revision}', default='',
                           help='The revision to initialize the blame file to')
    argparser.add_argument('--debug', action='store_true', default=False,
                           help='Increase logging and show tracebacks')
    argparser.add_argument('--version', action='version',
                           version='%s: %s\nPython: %s'
                           % (__name__, version.__version__, sys.version))
    argparser.add_argument('--disable-syntax', action='store_false',
                           default=True, dest='syntax',
                           help='Disable syntax highlighting')
    argparser.add_argument('filename', help='Name or path to file to blame')
    # These args are positional and prefixed with '+' to match vim.
    # Both are captured in the same argparse positional arg (must be positional
    # because arg parse doesn't allow options that start with a char other than
    # '-') and then it's parsed below
    argparser.add_argument('line_num_or_search_term', nargs='?',
                           metavar='+{line_num} or +/{search_pattern}',
                           help='The line number or search pattern the cursor '
                                'will be positioned on (this arg will put '
                                'bblame in visual mode)')
    args = argparser.parse_args()
    logging.info('Command line arguments: ' + str(args))
    if args.line_num_or_search_term:
        if args.line_num_or_search_term.startswith('+/'):
            # search term was passed in
            args.line_num_or_search_term = \
                args.line_num_or_search_term.strip('+/')
        elif args.line_num_or_search_term.startswith('+'):
            # Line number was passed in
            try:
                args.line_num_or_search_term = \
                    int(args.line_num_or_search_term.strip('+'))
            except ValueError as exc:
                # It certainly does have a message member, from baseException
                # pylint: disable=no-member
                if 'invalid literal for int' in exc.message:
                    sys.stderr.write('Invalid input for line number\n')
                    sys.exit(1)
        else:
            sys.stderr.write('Invalid positional argument: %s\n'
                             % args.line_num_or_search_term)
            sys.exit(1)
    try:
        os.stat(args.filename)
    except OSError as exc:
        if exc.errno == errno.ENOENT:
            sys.stderr.write('%s\n' % os.strerror(errno.ENOENT))
            exit(errno.ENOENT)
    if not os.path.isfile(args.filename):
        sys.stderr.write('Path provided does not lead to a file\n')
        exit(errno.ENOENT)
    return args


def main():
    """main callable that handles argument parsing the and then calls
    the curses_loop with the parsed arguments.
    Also sets up a few global settings and moves the cwd of the application
    to the appropriate location."""
    # reduce ESCAPE delay from 1s to 25ms
    os.environ['ESCDELAY'] = "25"
    # Process arguments passed from user
    args = process_args()

    # set logging level if debug
    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)

    abs_file_path = os.path.realpath(args.filename).strip()
    # Move cwd to git root dir, whichever git repo the input path leads us to,
    # to simplify git commands later.
    # Change to base dir of path provided
    os.chdir(os.path.dirname(abs_file_path))
    # Filenames may be relative to a sub dir (bblame ../../tests/mytestfile)
    # so recreate the path relative to the git root (tests/mytestfile) since
    # the application is run from git root.
    try:
        # pylint: disable=too-many-function-args
        git_root = sh.git('rev-parse', '--show-toplevel').stdout.strip()
        args.filename = abs_file_path.replace(git_root.decode('UTF-8'),
                                              '').strip('/')
        os.chdir(git_root)

        # Enter curses loop
        curses.wrapper(curses_loop, args)
    except (sh.ErrorReturnCode_128, git.BadRevException,
            git.NoSuchRevException) as exc:
        logging.debug(traceback.format_exc())
        sys.stderr.write(str(exc))
        sys.exit(1)


if __name__ == "__main__":
    main()



"""Setup script for building/installing package using distutils"""
from os import path
from setuptools import setup

PROG_NAME = 'bblame'
PACKAGE_NAME = 'betterblame'
# Use README as the long description
PROJECT_DIR = path.abspath(path.dirname(__file__))
with open(path.join(PROJECT_DIR, 'README.rst')) as f:
    LONG_DESCRIPTION = f.read()

# Load the __version__ string into the setup.py namespace with exec instead of
# importing, since there are known issues with the importing approach wreaking
# havoc in projects that have dependencies
__version__ = '0.0.0'  # default value and makes pylint happy
version_module = '%s/version.py' % PACKAGE_NAME
exec(open(version_module).read())

setup(name=PROG_NAME,
      # pylint: disable=undefined-variable
      version=__version__,
      description='An ncurses app for browsing file git history',
      long_description=LONG_DESCRIPTION,
      url='https://bitbucket.org/niko333/betterblame',
      author='Niko Oliveira',
      author_email='oliveira.n3@gmail.com',
      license='MIT',
      classifiers=['Development Status :: 4 - Beta',
                   'Intended Audience :: Developers',
                   'Topic :: Software Development :: Build Tools',
                   'Environment :: Console :: Curses',
                   'License :: OSI Approved :: MIT License',
                   'Programming Language :: Python :: 3'],
      keywords=['better git blame history curses ncurses'],
      packages=[PACKAGE_NAME],
      install_requires=[
          'texteditpad',
          'pygments',
          'future',
          'sh>=2',
          'six',
      ],
      scripts=[PROG_NAME],
      # entry_points={
      #     'console_scripts': [
      #         'bblame = bblame.bblame:main',
      #     ]
      # },
      zip_safe=False)



"""Utility functions and objects"""


def substring_split(start, end, string):
    """Split string at start:end, returning all three substrings (beginning,
    middle, end) if available"""
    # print('substring_split, start: %d, end: %d' % (start, end))
    return string[:start], string[start:end], string[end:]


def left_str_strip(string, str_to_strip):
    """Strips whole string <str_to_strip> from <string> from the left side.
    Distinct from lstrip, since the latter strips all chars, in any order, from
    the string"""
    return string[len(str_to_strip):]


def check_string(string, msg):
    """Checks if <string> is a string type in a python 2 and 3 compatible way.
    Raises <msg> in the AssertionError """
    try:
        basestring  # pylint: disable=E0601
    except NameError:
        basestring = str
    assert isinstance(string, basestring), msg


# In this case I really do want an object pylint, but thanks...
# pylint: disable=R0903
class BidirectionalCycle(object):
    """A cycle iterator that can iterate in both directions (e.g. has next
    and prev).
    This is a simple object that supports the iterator protocol but it doesn't
    behave like one might expect a standard iterator to (e.g. a generator that
    lazily produces the next value) this object will keep the WHOLE LIST in
    memory, so use WITH CAUTION
    >>> bi_iter = BidirectionalCycle([0, 1, 2])
    >>> bi_iter.next()
    0
    >>> bi_iter.next()
    1
    >>> bi_iter.next()
    2
    >>> bi_iter.next()
    0
    >>> bi_iter.prev()
    2
    >>> bi_iter.prev()
    1
    >>> bi_iter.prev()
    0
    >>> bi_iter.prev()
    2
    >>> bi_iter = BidirectionalCycle([0, 1, 2], starting_index=1)
    >>> bi_iter.next()
    1
    >>> bi_iter = BidirectionalCycle([0, 1, 2], starting_index=1)
    >>> bi_iter.prev()
    1
    >>> bi_iter = BidirectionalCycle([0, 1, 2], starting_index=1, no_wrap=True)
    >>> bi_iter.next()
    1
    >>> bi_iter.next()
    2
    >>> bi_iter.next()
    Traceback (most recent call last):
    ...
    StopIteration
    """
    def __init__(self, list_seq, starting_index=0, no_wrap=False):
        self.current_index = self.init_index = starting_index
        # CURRENTLY ONLY SUPPORT LISTS
        assert isinstance(list_seq, list), 'Currently only supports lists'
        self.seq = list_seq
        self.no_wrap = no_wrap
        self.start_of_day = True

    def next(self):
        """Maintain support for python2 iterator"""
        return self.__next__()

    def __next__(self):
        """return the next item in the iteration"""
        self._check_len()
        if self.start_of_day:
            return self._start_of_day()

        self._move_index_next()
        next_item = self.seq[self.current_index]

        return next_item

    def prev(self):
        """return the previous item in the iteration"""
        self._check_len()
        if self.start_of_day:
            if self.no_wrap:
                raise StopIteration()
            return self._start_of_day()

        self._move_index_prev()
        prev_item = self.seq[self.current_index]

        return prev_item

    def curr(self):
        """Returns the current item in the iteration"""
        self._check_len()
        if self.start_of_day:
            return self._start_of_day()

        return self.seq[self.current_index]

    def _move_index_next(self):
        """Move the index in the next direction"""
        # check if we need to wrap around to the beginning
        if self.current_index == len(self.seq) - 1:
            if self.no_wrap:
                raise StopIteration()
            self.current_index = 0
        else:
            self.current_index = self.current_index + 1

    def _move_index_prev(self):
        """Move the index in the prev direction"""
        # check if we need to wrap around to the end
        if self.current_index == 0:
            if self.no_wrap:
                raise StopIteration()
            self.current_index = len(self.seq) - 1
        else:
            self.current_index = self.current_index - 1

    def _check_len(self):
        """As itertools.cycle does, raise StopIteration if the sequence
        is empty"""
        if len(self.seq) == 0:
            raise StopIteration

    def _start_of_day(self):
        """print out the init_index of the sequence and set start_of_day to
        false. This is needed to get the behaviour that after init-ing the
        iterator if you call either previous or next, you get the starting
        index."""
        self.start_of_day = False
        return self.seq[self.init_index]

    def __str__(self):
        return str(self.seq)

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.seq)

    def __contains__(self, item):
        return item in self.seq
